

import os
from six.moves.urllib.parse import urlparse
from twisted import logger
from twisted.internet.task import deferLater
from twisted.internet.defer import CancelledError

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from kivy.uix.relativelayout import RelativeLayout

from .basemixin import BaseMixin
from .basemixin import BaseGuiMixin
from .resources import ASSET
from .widgets.gallery import ImageGallery
from .widgets.pdfplayer import PDFPlayer

WEBRESOURCE = 1

SIDEBAR = 1
PLAYER = 2

Base = declarative_base()
metadata = Base.metadata


class WebResourceGalleryModel(Base):
    __tablename__ = 'gallery_1'

    id = Column(Integer, primary_key=True)
    seq = Column(Integer, unique=True, index=True)
    rtype = Column(Integer)
    resource = Column(Text, index=True)
    duration = Column(Integer)

    def __repr__(self):
        return "{0:3} {1} {2} [{3}]".format(
            self.seq, self.rtype, self.resource, self.duration or ''
        )


class GalleryResource(object):
    def __init__(self, manager, seq=None, rtype=None,
                 resource=None, duration=None):
        self._manager = manager
        self._seq = seq

        self._rtype = None
        self.rtype = rtype

        self._resource = None
        self.resource = resource

        self._duration = duration

        if not self._rtype:
            self.load()

    @property
    def seq(self):
        return self._seq

    @property
    def rtype(self):
        return self._rtype

    @rtype.setter
    def rtype(self, value):
        if value not in [None, WEBRESOURCE]:
            raise ValueError
        self._rtype = value

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, value):
        if self._rtype == WEBRESOURCE:
            self._resource = os.path.basename(urlparse(value).path)

    @property
    def duration(self):
        return self._duration

    def commit(self):
        session = self._manager.db()
        try:
            try:
                robj = session.query(self._db_model).filter_by(seq=self.seq).one()
            except NoResultFound:
                robj = self._db_model()
                robj.seq = self.seq

            robj.rtype = self.rtype
            robj.resource = self.resource
            robj.duration = self._duration

            session.add(robj)
            session.commit()
            session.flush()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def load(self):
        session = self._manager.db()
        try:
            robj = session.query(self._db_model).filter_by(seq=self._seq).one()
            self.rtype = robj.rtype
            self.resource = robj.resource
            self._duration = robj.duration
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @property
    def _db_model(self):
        return self._manager.db_model


class BaseGalleryManager(object):
    def __init__(self, node, gmid, widget, default_duration=4):
        self._gmid = gmid
        self._log = None
        self._node = node
        self._widget = widget
        self._default_duration = default_duration
        self._seq = 0
        self._task = None
        self._items = []

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="gallery.{0}".format(self._gmid),
                                      source=self)
        return self._log

    @property
    def default_duration(self):
        return self._default_duration

    def flush(self, force=False):
        self.log.debug("Flushing gallery resources")
        self._items = []
        if force:
            self._trigger_transition()

    def _load(self, items):
        return items

    def load(self, items):
        self.log.debug("Loading gallery resource list")
        self.flush()
        self._items = self._load(items)

    def add_item(self, item):
        raise NotImplementedError

    def remove_item(self, item):
        raise NotImplementedError

    @property
    def current_seq(self):
        return self._seq

    @property
    def next_seq(self):
        seq = self._seq + 1
        if seq < len(self._items):
            return seq
        elif len(self._items):
            return 0
        else:
            return -1

    def start(self):
        self.log.info("Starting Gallery Manager {gmid} of {name}",
                            gmid=self._gmid, name=self.__class__.__name__)
        self.step()

    def step(self):
        self._seq = self.next_seq
        duration = self._trigger_transition(stopped=False)
        if not duration:
            duration = self.default_duration
        self._task = deferLater(self._node.reactor, duration, self.step)

        def _cancel_handler(failure):
            failure.trap(CancelledError)
        self._task.addErrback(_cancel_handler)
        return self._task

    def _trigger_transition(self, stopped=False):
        # If current_seq is -1, that means the gallery is empty. This may be
        # called repeatedly with -1. Use the returned duration to slow down
        # unnecessary requests. This function should also appropriately handle
        # creating or destroying gallery components.
        if stopped or self.current_seq == -1:
            self._widget.current = None
            return 30
        target = self._items[self.current_seq]
        duration = target.duration
        if target.rtype == WEBRESOURCE:
            fp = self._node.resource_manager.get(target.resource).filepath

            if not os.path.exists(fp):
                self._widget.current = None
                return 10

            if os.path.splitext(fp)[1] == '.pdf':
                fp = PDFPlayer(source=fp, exit_retrace=True,
                               temp_dir=self._node.temp_dir)
                if not target.duration:
                    duration = fp.num_pages * fp.interval

            self._widget.current = fp
        return duration

    def stop(self):
        if self._task:
            self._task.cancel()
        self._trigger_transition(stopped=True)

    def render(self):
        for item in self._items:
            print(item)


class GalleryManager(BaseGalleryManager):
    def __init__(self, *args, **kwargs):
        super(GalleryManager, self).__init__(*args, **kwargs)

        self._db_engine = None
        self._db = None
        self._db_dir = None
        _ = self.db

        self._persistence_load()

    def flush(self, force=False):
        super(GalleryManager, self).flush(force=force)
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        except NoResultFound:
            session.close()
            return
        try:
            for robj in results:
                session.delete(robj)
                # Orphan the resource so that the cache infrastructure
                # will clear the files as needed
                r = self._node.resource_manager.get(robj.resource)
                r.rtype = None
                r.commit()
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def _load(self, items):
        _items = []
        for idx, (resource, duration) in enumerate(items):
            robj = GalleryResource(
                self, seq=idx, rtype=WEBRESOURCE,
                resource=resource, duration=duration
            )
            robj.commit()
            _items.append(robj)
            r = self._node.resource_manager.get(resource)
            r.rtype = ASSET
            r.commit()
        self._fetch()
        return _items

    def _persistence_load(self):
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        except NoResultFound:
            session.close()
            return
        try:
            _items = []
            for robj in results:
                _items.append(GalleryResource(self, robj.seq))
        finally:
            session.close()
        self._items = _items
        self._fetch()

    def db_get_resources(self, session, seq=None):
        q = session.query(self.db_model)
        if seq is not None:
            q = q.filter(
                self.db_model.seq == seq
            )
        else:
            q = q.order_by(self.db_model.seq)
        return q

    @property
    def db_model(self):
        if self._gmid == SIDEBAR:
            return WebResourceGalleryModel

    @property
    def db(self):
        if self._db is None:
            self._db_engine = create_engine(self.db_url)
            metadata.create_all(self._db_engine)
            self._db = sessionmaker(expire_on_commit=False)
            self._db.configure(bind=self._db_engine)
        return self._db

    @property
    def db_url(self):
        return 'sqlite:///{0}'.format(os.path.join(self.db_dir, 'gallery.db'))

    @property
    def db_dir(self):
        return self._node.db_dir

    def _fetch(self):
        self.log.debug("Triggering Gallery Fetch")
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()
        for e in results:
            r = self._node.resource_manager.get(e.resource)
            self._node.resource_manager.prefetch(
                r, semaphore=self._node.http_semaphore_download
            )

    def render(self):
        print("----------")
        super(GalleryManager, self).render()
        print("DB Content")
        print("----------")
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        finally:
            session.close()
        print(results)
        print("----------")


class GalleryMixin(BaseMixin):
    def __init__(self, *args, **kwargs):
        self._gallery_managers = {}
        super(GalleryMixin, self).__init__(*args, **kwargs)

    def gallery_manager(self, gmid):
        if gmid not in self._gallery_managers.keys():
            self.log.info("Initializing gallery manager {gmid}", gmid=gmid)
            self._gallery_managers[gmid] = GalleryManager(self, gmid, self.gui_gallery)
        return self._gallery_managers[gmid]

    def gallery_load(self, items):
        self.gallery_manager(SIDEBAR).load(items)

    def gallery_start(self):
        self.gallery_manager(SIDEBAR).start()

    def gallery_stop(self):
        self.gallery_manager(SIDEBAR).stop()

    @property
    def gui_gallery(self):
        raise NotImplementedError


class GalleryGuiMixin(GalleryMixin, BaseGuiMixin):
    _media_extentions_image = ['.png', '.jpg', '.bmp', '.gif', '.jpeg']

    def __init__(self, *args, **kwargs):
        self._gallery = None
        self._gallery_parent_layout = None
        super(GalleryGuiMixin, self).__init__(*args, **kwargs)

    @property
    def gui_gallery_parent(self):
        if not self._gallery_parent_layout:
            self._gallery_parent_layout = RelativeLayout()
            self.gui_sidebar_right.add_widget(self._gallery_parent_layout)
        return self._gallery_parent_layout

    def _gui_gallery_sidebar_control(self, *args):
        if self.gui_gallery.visible:
            self.gui_sidebar_right_show('gallery')
        else:
            self.gui_sidebar_right_hide('gallery')

    @property
    def gui_gallery(self):
        if not self._gallery:
            self._gallery = ImageGallery(parent_layout=self.gui_gallery_parent)
            self._gallery.bind(visible=self._gui_gallery_sidebar_control)
        return self._gallery

    def gui_setup(self):
        super(GalleryGuiMixin, self).gui_setup()





from kivy.uix.label import Label
from .colors import BackgroundColorMixin


class ColorLabel(BackgroundColorMixin, Label):
    def __init__(self, **kwargs):
        bgcolor = kwargs.pop('bgcolor', None)
        Label.__init__(self, **kwargs)
        BackgroundColorMixin.__init__(self, bgcolor=bgcolor)


class SelfScalingLabel(Label):
    def __init__(self, **kwargs):
        kwargs['max_lines'] = 1
        super(SelfScalingLabel, self).__init__(**kwargs)
        self.bind(texture_size=self._scale_font)

    def _scale_font(self, *_):
        # print("Scaling {0} ?> {1}".format(self.texture_size[0], self.width))
        if self.texture_size[0] > self.width:
            self.font_size -= 1  # reduce font size if too wide


class SelfScalingColorLabel(SelfScalingLabel, ColorLabel):
    def __init__(self, **kwargs):
        super(SelfScalingColorLabel, self).__init__(**kwargs)





import os
from kivy.core.text import FontContextManager

from hoshi.i18n import TranslationManager

from .basemixin import BaseMixin
from .config import ConfigMixin
from .log import NodeLoggingMixin
from .basemixin import BaseGuiMixin


class AdvancedTextMixin(NodeLoggingMixin, ConfigMixin, BaseMixin):
    _supported_languages = ['en_US']

    def __init__(self, *args, **kwargs):
        super(AdvancedTextMixin, self).__init__(*args, **kwargs)
        self._i18n = TranslationManager(self.i18n_supported_languages,
                                        self._i18n_catalog_dirs)

    @property
    def _i18n_catalog_dirs(self):
        return [os.path.join(x, 'locale') for x in self.config.roots]

    @property
    def i18n_supported_languages(self):
        """
        Return the list of languages supported by the application. The list contains
        locale codes of the form 'en_US'. This is largely intended for use by code
        which requires i18n support. Such code should use this list to create and
        install a suitable set of i18n_contexts.
        """
        return self._supported_languages

    @property
    def i18n(self):
        return self._i18n

    def install(self):
        super(AdvancedTextMixin, self).install()
        self._i18n.install()


class AdvancedTextGuiMixin(AdvancedTextMixin, BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._text_font_context = None
        super(AdvancedTextGuiMixin, self).__init__(*args, **kwargs)

    @property
    def text_font_context(self):
        if not self._text_font_context and self.config.text_use_fcm:
            self._text_create_fcm()
        return self._text_font_context

    def _text_create_fcm(self):
        fc = self._appname
        if self.config.text_fcm_system:
            fc = "system://{0}".format(fc)
        self._text_font_context = fc
        self.log.info("Creating FontContextManager {0} using fonts in {1}"
                      .format(fc, self.config.text_fcm_fonts))
        FontContextManager.create(fc)

        for filename in os.listdir(self.config.text_fcm_fonts):
            self.log.debug("Installing Font {0} to FCM {1}".format(filename, self._text_font_context))
            FontContextManager.add_font(fc, os.path.join(self.config.text_fcm_fonts, filename))

    @property
    def text_font_params(self):
        params = {}
        if self.text_font_context:
            params.update({
                'font_context': self._text_font_context
            })
        else:
            params.update({
                'font_name': self.config.text_font_name
            })
        return params





class CompositeAnimationManager(object):
    def __init__(self):
        self._animations = []
        self._finish_handler = None

    def __len__(self):
        return len(self._animations)

    def add(self, animation, widget):
        self._animations.append((animation, widget))

    def _anim_done_handler(self, anim, widget):
        anim.cancel(widget)
        self._animations.remove((anim, widget))
        if len(self._animations) == 0:
            self._finish_handler()

    def when_done(self, handler):
        self._finish_handler = handler

    def start(self):
        for animation, widget in self._animations:
            animation.bind(on_complete=self._anim_done_handler)
            animation.start(widget)

    def cancel(self):
        for anim, widget in self._animations:
            anim.cancel_all(widget)

    def clear(self):
        self.cancel()
        self._animations = []




from twisted.internet.task import deferLater
from twisted.internet.defer import CancelledError
from .labels import SelfScalingLabel

import arrow


class ClockBase(object):
    def __init__(self, node):
        self._node = node
        self._update_task = None
        super(ClockBase, self).__init__()

    def update(self):
        raise NotImplementedError

    def start(self):
        self._node.log.info("Starting Update Task for {0}".format(self))
        self.step()

    def step(self):
        self.update()
        self._update_task = deferLater(self._node.reactor, 1, self.step)

        def _cancel_handler(failure):
            failure.trap(CancelledError)

        self._update_task.addErrback(_cancel_handler)

        return self._update_task

    def stop(self):
        self._node.log.info("Stopping Update Task for {0}".format(self))
        if self._update_task:
            self._update_task.cancel()


class SimpleDigitalClock(ClockBase, SelfScalingLabel):
    def __init__(self, node, **kwargs):
        ClockBase.__init__(self, node)
        SelfScalingLabel.__init__(self, **kwargs)

    def update(self):
        self.text = arrow.now().format("HH:mm:ss")





from .log import LoggingGuiMixin
from .nodeid import NodeIDGuiMixin
from .busy import BusySpinnerGuiMixin
from .background import OverlayWindowGuiMixin
from .marquee import MarqueeGuiMixin
from .text import AdvancedTextGuiMixin

from .log import NodeLoggingMixin
from .nodeid import NodeIDMixin
from .busy import NodeBusyMixin
from .http import HttpClientMixin
from .shell import BaseShellMixin
from .text import AdvancedTextMixin

from .resources import ResourceManagerMixin


class BaseIoTNode(ResourceManagerMixin, HttpClientMixin, BaseShellMixin,
                  NodeBusyMixin, AdvancedTextMixin, NodeLoggingMixin, NodeIDMixin):
    _has_gui = False

    def __init__(self, *args, **kwargs):
        super(BaseIoTNode, self).__init__(*args, **kwargs)

    def install(self):
        super(BaseIoTNode, self).install()
        self._log.info("Installing Node with ID {log_source.id}")

    def start(self):
        super(BaseIoTNode, self).start()
        self._log.info("Starting Node with ID {log_source.id}")

    def stop(self):
        super(BaseIoTNode, self).stop()
        self._log.info("Stopping Node with ID {log_source.id}")


class BaseIoTNodeGui(NodeIDGuiMixin, BusySpinnerGuiMixin, LoggingGuiMixin,
                     MarqueeGuiMixin, AdvancedTextGuiMixin, OverlayWindowGuiMixin,
                     BaseIoTNode):

    def __init__(self, *args, **kwargs):
        self._application = kwargs.pop('application')
        self._gui_root = None
        super(BaseIoTNodeGui, self).__init__(*args, **kwargs)

    @staticmethod
    def _gui_disable_multitouch_emulation():
        from kivy.config import Config
        Config.set('input', 'mouse', 'mouse,multitouch_on_demand')

    def gui_setup(self):
        self._gui_disable_multitouch_emulation()
        super(BaseIoTNodeGui, self).gui_setup()
        # # Setup GUI elements from other Mixins
        # OverlayWindowGuiMixin.gui_setup(self)
        # NodeIDGuiMixin.gui_setup(self)
        # LoggingGuiMixin.gui_setup(self)
        return self.gui_root





import os
from datetime import datetime
from datetime import timedelta
from cached_property import threaded_cached_property_with_ttl
from six.moves.urllib.parse import urlparse
from twisted.internet.task import deferLater
from twisted.internet.threads import deferToThread
from twisted.internet.defer import DeferredSemaphore
from twisted import logger

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy import DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from .basenode import BaseIoTNode
from .resources import CacheableResource
from .mediaplayer import MediaPlayerBusy
from .marquee import MarqueeBusy
from .widgets.pdfplayer import generate_pdf_images

from .constants import WEBRESOURCE
from .constants import TEXT

Base = declarative_base()
metadata = Base.metadata


class WebResourceEventsModel(Base):
    __tablename__ = 'events_1'

    id = Column(Integer, primary_key=True)
    eid = Column(Text, unique=True, index=True)
    etype = Column(Integer)
    resource = Column(Text, index=True)
    start_time = Column(DateTime, index=True)
    duration = Column(Text)
    
    def __repr__(self):
        return "{0:3} {2} {3:.2f} {1}".format(
            self.eid, self.resource, self.etype,
            (self.start_time - datetime.now()).total_seconds()
        )


class TextEventsModel(Base):
    __tablename__ = 'events_2'

    id = Column(Integer, primary_key=True)
    eid = Column(Text, unique=True, index=True)
    etype = Column(Integer)
    resource = Column(Text)
    start_time = Column(DateTime, index=True)
    duration = Column(Integer)
    
    def __repr__(self):
        return "{0:3} {2} {3:.2f} {1}".format(
            self.eid, self.resource, self.etype,
            (self.start_time - datetime.now()).total_seconds()
        )


class ScheduledResourceClass(CacheableResource):
    @threaded_cached_property_with_ttl(ttl=3)
    def next_use(self):
        next_event = self.node.event_manager(WEBRESOURCE).next(
            resource=self.filename)
        if next_event:
            return next_event.start_time
        else:
            return None


class Event(object):
    def __init__(self, manager, eid, etype=None, resource=None,
                 start_time=None, duration=None):
        self._manager = manager
        self._eid = eid

        self._etype = None
        self.etype = etype

        self._resource = None
        self.resource = resource

        self._start_time = None
        self.start_time = start_time

        self._duration = None
        self.duration = duration
        if not self._etype:
            self.load()

    @property
    def eid(self):
        return self._eid

    @property
    def etype(self):
        return self._etype

    @etype.setter
    def etype(self, value):
        if value not in [None, WEBRESOURCE, TEXT]:
            raise ValueError
        self._etype = value

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, value):
        if self.etype == WEBRESOURCE:
            self._resource = os.path.basename(urlparse(value).path)
        elif self.etype == TEXT:
            self._resource = value

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, value):
        if not value:
            return
        if isinstance(value, datetime):
            self._start_time = value
        else:
            self._start_time = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')

    @property
    def duration(self):
        return self._duration or None

    @duration.setter
    def duration(self, value):
        self._duration = int(value) if value else None

    def commit(self):
        session = self._manager.db()
        try:
            try:
                eobj = session.query(self._db_model).filter_by(eid=self.eid).one()
            except NoResultFound:
                eobj = self._db_model()
                eobj.eid = self.eid

            eobj.etype = self._etype
            eobj.resource = self.resource
            eobj.start_time = self.start_time
            eobj.duration = self.duration

            session.add(eobj)
            session.flush()
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def load(self):
        session = self._manager.db()
        try:
            eobj = session.query(self._db_model).filter_by(eid=self._eid).one()
            self.etype = eobj.etype
            self.resource = eobj.resource
            self._start_time = eobj.start_time
            self.duration = eobj.duration
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @property
    def _db_model(self):
        return self._manager.db_model

    def __repr__(self):
        return "{0:3} {2} {3:.2f} {1}".format(
            self.eid, self.resource, self.etype,
            (self.start_time - datetime.now()).total_seconds()
        )


class EventManager(object):
    def __init__(self, node, emid):
        self._emid = emid
        self._node = node
        self._db_engine = None
        self._db = None
        self._db_dir = None
        self._execute_task = None
        self._current_event = None
        self._current_event_resource = None
        self._preprocess_semaphore = None
        self._log = None
        _ = self.db

    @property
    def emid(self):
        return self._emid

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="em.{0}".format(self.emid), source=self)
        return self._log

    @property
    def preprocess_semaphore(self):
        if self._preprocess_semaphore is None:
            self._preprocess_semaphore = DeferredSemaphore(1)
        return self._preprocess_semaphore

    def insert(self, eid, **kwargs):
        event = Event(self, eid, **kwargs)
        event.commit()

    def remove(self, eid):
        session = self.db()
        # print("Trying to remove {0} from edb".format(eid))
        try:
            try:
                eobj = session.query(self.db_model).filter_by(eid=eid).one()
            except NoResultFound:
                return False
            # print("Commiting edel")
            session.delete(eobj)
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    def _pointers(self, cond, resource=None, follow=False):
        session = self.db()
        try:
            r = self.db_get_events(session, resource=resource).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()

        e = None
        l = None
        for e in r:
            if l and cond(l, e):
                break
            l = e
        if l:
            if not follow:
                return Event(self, l.eid)
            else:
                if e:
                    ne = Event(self, e.eid)
                else:
                    ne = None
                return Event(self, l.eid), ne
        if follow:
            return None, None
        else:
            return None

    def previous(self, resource=None, follow=False):
        return self._pointers(
            lambda l, e: e.start_time >= datetime.now(),
            resource=resource, follow=follow
        )

    def next(self, resource=None, follow=False):
        return self._pointers(
            lambda l, e: l.start_time >= datetime.now(),
            resource=resource, follow=follow
        )

    def get(self, eid):
        return Event(self, eid)

    def prune(self):
        # TODO This doesn't work!
        # with self.db as db:
        #     r = db[self.db_table_name].find(start_time={'lt': datetime.now()})
        #     for result in r:
        #         print("Removing {0}".format(r))
        #         self.remove(result['eid'])
        session = self.db()
        try:
            results = self.db_get_events(session).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()

        for result in results:
            if result.start_time >= datetime.now():
                break
            self.log.warn("Pruning missed event {event}", event=result)
            self.remove(result.eid)

    def render(self):
        session = self.db()
        try:
            results = self.db_get_events(session).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()
        for result in results:
            print(Event(self, result.eid))

    def db_get_events(self, session, resource=None):
        q = session.query(self.db_model)
        if resource:
            q = q.filter(
                self.db_model.resource == resource
            )
        q = q.order_by(self.db_model.start_time)
        return q

    @property
    def db_model(self):
        if self._emid == WEBRESOURCE:
            return WebResourceEventsModel
        elif self._emid == TEXT:
            return TextEventsModel

    @property
    def db(self):
        if self._db is None:
            self._db_engine = create_engine(self.db_url)
            metadata.create_all(self._db_engine)
            self._db = sessionmaker(expire_on_commit=False)
            self._db.configure(bind=self._db_engine)
        return self._db

    @property
    def db_url(self):
        return 'sqlite:///{0}'.format(os.path.join(self.db_dir, 'events.db'))

    @property
    def db_dir(self):
        return self._node.db_dir

    @property
    def current_event(self):
        return self._current_event
    
    @property
    def current_event_resource(self):
        return self._current_event_resource

    def _trigger_event(self, event):
        raise NotImplementedError

    def _finish_event(self, forced):
        if forced:
            self.log.info("Event {eid} was force stopped.",
                          eid=self._current_event)
        else:
            self.log.info("Successfully finished event {eid}",
                          eid=self._current_event)
            self._succeed_event(self._current_event)
        self._current_event = None
        self._current_event_resource = None

    def _succeed_event(self, event):
        raise NotImplementedError

    def _event_scheduler(self):
        event = None
        nevent = None
        le, ne = self.previous(follow=True)
        if le:
            ltd = datetime.now() - le.start_time
            # self.log.debug("S {emid} LTD {ltd}",
            #                ltd=ltd, emid=self._emid)
            if abs(ltd) < timedelta(seconds=3):
                event = le
                nevent = ne
        if not event:
            ne, nne = self.next(follow=True)
            if ne:
                ntd = ne.start_time - datetime.now()
                # self.log.debug("S {emid} NTD {ntd}",
                #                ntd=ntd, emid=self._emid)
                if abs(ntd) < timedelta(seconds=3):
                    event = ne
                    nevent = nne
        if event:
            retry = self._trigger_event(event)
            if retry:
                self._execute_task = deferLater(self._node.reactor, 0.1,
                                                self._event_scheduler)
                return
        self._execute_task = self._event_scheduler_hop(nevent)

    def _event_scheduler_hop(self, next_event=None):
        if not next_event:
            next_event = self.next()
        if not next_event:
            next_start = timedelta(seconds=60)
        else:
            next_start = next_event.start_time - datetime.now()
            # print("Next Start : ", next_start)
            if not next_event or next_start < timedelta(0):
                next_start = timedelta(seconds=60)
            elif next_start > timedelta(seconds=60):
                next_start = timedelta(seconds=60)
        self.log.debug("SCHED {emid} HOP {ns}", emid=self._emid,
                       ns=next_start.seconds)
        return deferLater(self._node.reactor, next_start.seconds,
                          self._event_scheduler)

    def start(self):
        self.log.info("Starting Event Manager {emid} of {name}",
                      emid=self._emid, name=self.__class__.__name__)
        self._event_scheduler()


class TextEventManager(EventManager):
    def _trigger_event(self, event):
        try:
            d = self._node.marquee_play(text=event.resource,
                                        duration=event.duration)
            d.addCallback(self._finish_event)
            self.log.info("Executed Event : {0}".format(event))
            self._current_event = event.eid
            self._current_event_resource = event.resource
        except MarqueeBusy as e:
            # self._node.log.warn("Marquee busy for {event} : {e}",
            #                     event=event, e=e.now_playing)
            return e.collision_count
        self.remove(event.eid)
        self.prune()

    def _succeed_event(self, event):
        try:
            self._node.api_text_success([event])
        except NotImplementedError:
            self.log.warn("Node has no text event success reporter")


class WebResourceEventManager(EventManager):
    def _trigger_event(self, event):
        r = self._node.resource_manager.get(event.resource)
        if r.available:
            try:
                d = self._node.media_play(content=r,
                                          duration=event.duration)
                d.addCallback(self._finish_event)
                self.log.info("Executed Event : {0}".format(event))
                self._current_event = event.eid
                self._current_event_resource = event.resource
            except MediaPlayerBusy as e:
                # self.log.warn("Mediaplayer busy for {event} : {e}",
                #               event=event, e=e.now_playing)
                return e.collision_count
        else:
            self.log.warn("Media not ready for {event}", event=event)
        self.remove(event.eid)
        self.prune()

    def _succeed_event(self, event):
        try:
            self._node.api_media_success([event])
        except NotImplementedError:
            self.log.warn("Node has no media event success reporter")

    def _preprocess_pdf(self, filepath):
        name = os.path.splitext(os.path.basename(filepath))[0]
        target = os.path.join(self._node.temp_dir, name)
        return deferToThread(generate_pdf_images, filepath, target, None)

    def _preprocess_resource(self, maybe_failure, resource):
        if os.path.splitext(resource.filename)[1] == '.pdf':
            self.preprocess_semaphore.run(
                self._preprocess_pdf, resource.filepath
            )

    def _fetch(self):
        self.log.debug("Triggering WebResource Fetch")
        session = self.db()
        try:
            results = self.db_get_events(session).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()
        for e in results:
            if e.start_time - datetime.now() > timedelta(seconds=1200):
                break
            r = self._node.resource_manager.get(e.resource)
            d = self._node.resource_manager.prefetch(
                r, semaphore=self._node.http_semaphore_download
            )
            d.addCallback(self._preprocess_resource, r)
        self._fetch_task = deferLater(self._node.reactor, 600, self._fetch)

    def _fetch_scheduler(self):
        self._fetch()

    def _prefetch(self):
        self.log.debug("Triggering WebResource Prefetch")
        session = self.db()
        try:
            results = self.db_get_events(session).all()
        except:
            session.rollback()
            raise
        finally:
            session.close()
        for e in results:
            if e.start_time - datetime.now() > timedelta(seconds=(3600 * 6)):
                break
            r = self._node.resource_manager.get(e.resource)
            self._node.resource_manager.prefetch(
                r, semaphore=self._node.http_semaphore_background
            )
        self._prefetch_task = deferLater(self._node.reactor, 3600, self._prefetch)

    def _prefetch_scheduler(self):
        self._prefetch()

    def start(self):
        super(WebResourceEventManager, self).start()
        self._fetch_scheduler()
        self._prefetch_scheduler()


class EventManagerMixin(BaseIoTNode):
    def __init__(self, *args, **kwargs):
        self._event_managers = {}
        self._success_api_engine = None
        super(EventManagerMixin, self).__init__(*args, **kwargs)

    def event_manager_install(self, manager):
        self.log.info("Installing Event Manager {1} with emid {0}".format(manager.emid, manager))
        self._event_managers[manager.emid] = manager

    def event_manager(self, emid):
        return self._event_managers[emid]

    @property
    def _cache_trim_exclusions(self):
        if self.event_manager(WEBRESOURCE).current_event_resource:
            return [self.event_manager(WEBRESOURCE).current_event_resource]
        else:
            return []

    def api_media_success(self, events):
        if not self._success_api_engine:
            raise NotImplementedError
        self._success_api_engine.api_media_success(events)

    def api_text_success(self, events):
        if not self._success_api_engine:
            raise NotImplementedError
        self._success_api_engine.api_text_success(events)

    def start(self):
        super(EventManagerMixin, self).start()
        for emid, manager in self._event_managers.items():
            manager.start()





import netifaces
import uuid

from twisted.internet import task

from .widgets.labels import ColorLabel
from .widgets.colors import color_set_alpha

from .basemixin import BaseMixin
from .basemixin import BaseGuiMixin

from .config import ConfigMixin


class NodeIDMixin(ConfigMixin, BaseMixin):
    _node_id_netifaces_fallback_interfaces = ['wlp1s0', 'wlan0', 'eth0']

    def __init__(self, *args, **kwargs):
        super(NodeIDMixin, self).__init__(*args, **kwargs)
        self._id = None

    @property
    def id(self):
        if self._id is None:
            self._id = self._get_id()
        return self._id

    def _get_id(self):
        if self.config.node_id_override is not None:
            return self.config.node_id_override
        getter = "_get_node_id_{0}".format(self.config.node_id_getter)
        params = {'interface': self.config.node_id_interface}
        return getattr(self, getter)(**params).upper()

    def _get_node_id_uuid(self, **_):
        node_id = uuid.getnode()
        if (node_id >> 40) % 2:
            raise OSError("The system does not seem to have a valid MAC")
        return hex(node_id)[2:]

    def _get_node_id_netifaces_guess(self):
        fallback_interfaces = self._node_id_netifaces_fallback_interfaces
        available_interfaces = netifaces.interfaces()
        default_gateway = netifaces.gateways()['default']
        if default_gateway:
            return default_gateway[netifaces.AF_INET][1]
        for iface in fallback_interfaces:
            if iface in available_interfaces:
                return iface

    def _get_node_id_netifaces(self, **kwargs):
        interface = kwargs.get('interface', None)
        if interface is None:
            interface = self._get_node_id_netifaces_guess()
        mac = netifaces.ifaddresses(interface)[netifaces.AF_LINK][0]['addr']
        return mac.replace(':', '')


class NodeIDGuiMixin(NodeIDMixin, BaseGuiMixin):
    _gui_nodeid_bgcolor = None
    _gui_nodeid_color = None

    def __init__(self, *args, **kwargs):
        super(NodeIDGuiMixin, self).__init__(*args, **kwargs)
        self._gui_id_tag = None
        self._gui_id_task = None

    @property
    def gui_id_tag(self):
        if not self._gui_id_tag:
            params = {'bgcolor': (self._gui_nodeid_bgcolor or
                                  color_set_alpha(self.gui_color_1, 0.4)),
                      'color': [1, 1, 1, 1]}
            self._gui_id_tag = ColorLabel(
                text=self.id, size_hint=(None, None),
                width=250, height=50, font_size='18sp',
                valign='middle', halign='center', **params
            )
        return self._gui_id_tag

    def gui_id_show(self, duration=None):
        if not self.gui_id_tag.parent:
            self.gui_status_stack.add_widget(self.gui_id_tag)
        if duration:
            self.reactor.callLater(duration, self.gui_id_hide)

    def gui_id_hide(self):
        if self.gui_id_tag.parent:
            self.gui_status_stack.remove_widget(self.gui_id_tag)

    def gui_setup(self):
        super(NodeIDGuiMixin, self).gui_setup()
        if not self.config.node_id_display:
            return
        if self.config.node_id_display_frequency:
            self._gui_id_task = task.LoopingCall(
                self.gui_id_show,
                duration=self.config.node_id_display_duration
            )
            self._gui_id_task.start(self.config.node_id_display_frequency)
        else:
            self.gui_id_show(duration=self.config.node_id_display_duration)





import arrow
from twisted import logger
from twisted.internet.task import deferLater
from twisted.internet.defer import CancelledError

from ebs.iot.linuxnode.log import NodeLoggingMixin
from ebs.iot.linuxnode.text import AdvancedTextMixin
from ebs.iot.linuxnode.text import AdvancedTextGuiMixin
from ebs.iot.linuxnode.basemixin import BaseGuiMixin

from ebs.iot.linuxnode.tables.animated import AnimatedTable
from ebs.iot.linuxnode.tables.renderable import BasicRenderableTableEntry
from ebs.iot.linuxnode.tables.spec import BasicTablePalette
from ebs.iot.linuxnode.tables.spec import BasicTableSpec
from ebs.iot.linuxnode.tables.spec import BasicColumnSpec


class TimetableEntry(BasicRenderableTableEntry):
    def __init__(self, data):
        super(TimetableEntry, self).__init__(data)

    @property
    def name(self):
        raise NotImplementedError

    @property
    def ts_start(self):
        raise NotImplementedError

    @property
    def ts_end(self):
        raise NotImplementedError


class Timetable(AnimatedTable):
    _prior_window = 30
    _post_window = 60
    _period_page = 12

    def __init__(self, node, spec=None):
        self._log = None
        self._redraw_task = None
        self._current_page = 0
        if not spec:
            spec = BasicTableSpec(self, [
                BasicColumnSpec("Name", 'name'),
                BasicColumnSpec("Start Time", "ts_start"),
                BasicColumnSpec("End Time", "ts_end")],
                show_column_header=True,
                row_height=90, row_spacing=10, font_size='42sp', font_bold=False)
        super(Timetable, self).__init__(node, spec)

    @property
    def prior_window(self):
        return self._prior_window

    @property
    def post_window(self):
        return self._post_window

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="timetable.{0}".format(self._spec.name),
                                      source=self)
        return self._log

    def build(self, entries):
        return super(Timetable, self).build(entries=entries)

    @property
    def active_entries(self):
        return [x for x in sorted(self._entries, key=lambda y: y.ts_start)
                if x.ts_start.shift(minutes=-1 * self.post_window) < arrow.now() < x.ts_end.shift(minutes=self.prior_window)]

    def start(self):
        self.log.info("Starting Timetable Redraw Task for {0}".format(self))
        self._current_page = 0
        self.step()

    def _turn_page(self):
        self._current_page += 1
        if self._current_page >= self.total_pages:
            self.next_language()
            self.log.debug("Switched to Next Language : {}".format(self._i18n_language))
            self._current_page = 0

    def step(self):
        self.log.debug("Drawing page {0} / {1}".format(self._current_page + 1, self.total_pages))
        self._turn_page()
        self.redraw_entries(entries=self.page_entities(self._current_page))

        duration = self._period_page
        self._redraw_task = deferLater(self._node.reactor, duration, self.step)

        def _cancel_handler(failure):
            failure.trap(CancelledError)
        self._redraw_task.addErrback(_cancel_handler)

        return self._redraw_task

    def retrigger(self):
        if self._redraw_task.active():
            self._redraw_task.cancel()
        self.step()

    def stop(self):
        self.log.info("Stopping Timetable Redraw Task for {0}".format(self))
        if self._redraw_task:
            self._redraw_task.cancel()


class BaseTimetableMixin(AdvancedTextMixin, NodeLoggingMixin):
    _timetable_class = Timetable
    _timetable_entry_class = TimetableEntry

    def __init__(self, *args, **kwargs):
        self._timetable = None
        super(BaseTimetableMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(BaseTimetableMixin, self).install()
        self.log.info("Installing TimeTable {0} with {1}"
                      "".format(self._timetable_class.__name__,
                                self._timetable_entry_class.__name__))
        self._timetable = self._timetable_class(self)

    def timetable_update(self, data, incremental=False):
        self._timetable.update([self._timetable_entry_class(x) for x in data],
                               incremental)


class BaseTimetableGuiMixin(AdvancedTextGuiMixin, BaseTimetableMixin, BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._timetable_palette = None
        self._timetable_gui = None
        self._task_timetable_redraw = None
        super(BaseTimetableGuiMixin, self).__init__(*args, **kwargs)

    @property
    def timetable_palette(self):
        if not self._timetable_palette:
            self._timetable_palette = BasicTablePalette(
                color_cell_background=self.gui_color_2,
                color_cell_foreground=self.gui_color_foreground,
                color_header_cell_background=self.gui_color_1,
                color_header_cell_foreground=self.gui_color_foreground,
                color_grid_background=self.gui_color_background,
            )
        return self._timetable_palette

    @property
    def timetable_gui(self):
        if not self._timetable_gui:
            self._timetable.palette = self.timetable_palette
            self._timetable_gui = self._timetable.build(entries=self._timetable.active_entries)
            self._timetable.start()
        return self._timetable_gui

    def _timetable_redraw(self):
        self._timetable.build(entries=self._timetable.active_entries)





import os
import subprocess

from kivy.core.window import Window
from kivy.uix.video import Video
from kivy.uix.boxlayout import BoxLayout
from six.moves.urllib.parse import urlparse

from .widgets.image import BleedImage
from .basemixin import BaseGuiMixin
from .config import ConfigMixin
from .externalplayer import ExternalMediaPlayer


class BackgroundGuiMixin(ConfigMixin, BaseGuiMixin):
    _bg_structured_prefix = 'structured:'
    _bg_separator = ':'

    def __init__(self, *args, **kwargs):
        self._bg_image = None
        self._bg_video = None
        self._bg_structured = None
        self._bg = None
        self._bg_container = None
        self._bg_current = None
        super(BackgroundGuiMixin, self).__init__(*args, **kwargs)

    def bg_is_structured(self, value):
        return value.startswith(self._bg_structured_prefix)

    def bg_is_file(self, value):
        return not value.startswith(self._bg_structured_prefix)

    def background_set(self, fpath):
        if not fpath:
            fpath = None

        if self.bg_is_structured(fpath):
            if not hasattr(self, fpath.split(self._bg_separator)[1]):
                fpath = None
        else:
            if not os.path.exists(fpath):
                fpath = None

        if self.config.background != fpath:
            old_bg = os.path.basename(urlparse(self.config.background).path)
            if self.bg_is_file(old_bg) and self.resource_manager.has(old_bg):
                self.resource_manager.remove(old_bg)
            self.config.background = fpath

        self.gui_bg_update()

    @property
    def gui_bg_container(self):
        if self._bg_container is None:
            self._bg_container = BoxLayout()
            self.gui_main_content.add_widget(self._bg_container)

            def _child_geometry(widget, _):
                if isinstance(self._bg, ExternalMediaPlayer):
                    self._bg.set_geometry(
                        widget.x, widget.y, widget.width, widget.height
                    )
            self.gui_main_content.bind(size=_child_geometry,
                                       pos=_child_geometry)
        return self._bg_container

    def gui_bg_clear(self):
        if self._bg_image:
            self.gui_bg_container.remove_widget(self._bg_image)
            self._bg_image = None
        if self._bg_structured:
            if hasattr(self._bg_structured, 'stop'):
                self._bg_structured.stop()
            self.gui_bg_container.remove_widget(self._bg_structured)
            self._bg_structured = None
        if self._bg_video:
            if isinstance(self._bg_video, Video):
                self.gui_bg_container.remove_widget(self._bg_video)
                self._bg_video.unload()
            elif isinstance(self._bg_video, ExternalMediaPlayer):
                self._bg_video.force_stop()
            self._bg_video = None
        self._bg = None

    @property
    def gui_bg_image(self):
        return self._bg_image

    @gui_bg_image.setter
    def gui_bg_image(self, value):
        if not os.path.exists(value):
            return

        self.gui_bg_clear()

        self._bg_image = BleedImage(
            source=value,
            allow_stretch=True,
            keep_ratio=True,
            bgcolor='auto'
        )
        self._bg = self._bg_image
        self.gui_bg_container.add_widget(self._bg_image)

    @property
    def gui_bg_video(self):
        return self._bg_video

    def _gui_bg_video_native(self, value):
        self._bg_video = Video(
            source=value, state='play',
            allow_stretch=True,
        )

        def _when_done(*_):
            self._bg_video.state = 'play'

        self._bg_video.bind(eos=_when_done)
        self._bg = self._bg_video
        self.gui_bg_container.add_widget(self._bg_video)

    def _gui_bg_video_external(self, value):
        self.log.info("Starting external background video player")
        geometry = (self.gui_bg_container.x, self.gui_bg_container.y,
                    self.gui_bg_container.width, self.gui_bg_container.height)
        self._bg_video = ExternalMediaPlayer(
            value, geometry, None, self, loop=True,
            layer=self.config.background_dispmanx_layer,
            dbus_name='org.mpris.MediaPlayer2.omxplayer1'
        )

        self._bg = self._bg_video
        self._bg_external = self._bg_video

    @gui_bg_video.setter
    def gui_bg_video(self, value):
        if not os.path.exists(value):
            return

        self.gui_bg_clear()

        if self.config.background_external_player:
            self._gui_bg_video_external(value)
        else:
            self._gui_bg_video_native(value)

    @property
    def gui_bg_structured(self):
        return self._bg_structured

    @gui_bg_structured.setter
    def gui_bg_structured(self, value):
        if not hasattr(self, value):
            return

        self.gui_bg_clear()

        self._bg_structured = getattr(self, value)
        self._bg = self._bg_structured
        self.gui_bg_container.add_widget(self._bg_structured)


    @property
    def gui_bg(self):
        return self._bg

    @gui_bg.setter
    def gui_bg(self, value):
        self.log.info("Setting background to {value}", value=value)
        if self.bg_is_file(value) and not os.path.exists(value):
            value = self.config.background

        if value == self._bg_current:
            return

        if self.bg_is_file(value):
            _media_extentions_image = ['.png', '.jpg', '.bmp', '.gif', '.jpeg']
            if os.path.splitext(value)[1] in _media_extentions_image:
                self.gui_bg_image = value
            else:
                self.gui_bg_video = value
        else:
            self.gui_bg_structured = value.split(self._bg_separator)[1]
        self._bg_current = value

    def gui_bg_update(self):
        self.gui_bg = self.config.background

    def gui_bg_pause(self):
        self.log.debug("Pausing Background")
        self.gui_main_content.remove_widget(self._bg_container)
        if isinstance(self.gui_bg, Video):
            self.gui_bg.state = 'pause'
        elif isinstance(self.gui_bg, ExternalMediaPlayer):
            self.gui_bg.pause()

    def gui_bg_resume(self):
        self.log.debug("Resuming Background")
        if not self._bg_container.parent:
            self.gui_main_content.add_widget(self._bg_container, len(self.gui_main_content.children))
        if hasattr(self.gui_bg, 'retrigger'):
            self.gui_bg.retrigger()
        if isinstance(self.gui_bg, Video):
            self.gui_bg.state = 'play'
        elif isinstance(self.gui_bg, ExternalMediaPlayer):
            self.gui_bg.resume()

    def stop(self):
        if self._bg and isinstance(self._bg, ExternalMediaPlayer):
            self._bg.force_stop()
        super(BackgroundGuiMixin, self).stop()

    def gui_setup(self):
        super(BackgroundGuiMixin, self).gui_setup()
        self.gui_bg = self.config.background


class OverlayWindowGuiMixin(BackgroundGuiMixin):
    # Overlay mode needs specific host support.
    # RPi :
    #   See DISPMANX layers and
    #   http://codedesigner.de/articles/omxplayer-kivy-overlay/index.html
    # Normal Linux Host :
    #   See core-x11 branch and
    #   https://groups.google.com/forum/#!topic/kivy-users/R4aJCph_7IQ
    # Others :
    #   Unknown, see
    #   - https://github.com/kivy/kivy/issues/4307
    #   - https://github.com/kivy/kivy/pull/5252
    _gui_supports_overlay_mode = False

    def __init__(self, *args, **kwargs):
        self._overlay_mode = None
        self._foundation_process = None
        super(OverlayWindowGuiMixin, self).__init__(*args, **kwargs)

    @property
    def overlay_mode(self):
        return self._overlay_mode

    @overlay_mode.setter
    def overlay_mode(self, value):
        if not self._gui_supports_overlay_mode:
            self.log.warn("Application tried to change overlay mode, "
                          "not supported this platform.")
            return
        if value is True:
            self._gui_overlay_mode_enter()
        else:
            self._gui_overlay_mode_exit()

    def _gui_overlay_mode_enter(self):
        self.log.info('Entering Overlay Mode')
        if self._overlay_mode:
            return
        self._overlay_mode = True
        Window.clearcolor = [0, 0, 0, 0]
        # self.gui_bg_pause()

    def _gui_overlay_mode_exit(self):
        self.log.info('Exiting Overlay Mode')
        if not self._overlay_mode:
            return
        self._overlay_mode = False
        # self.gui_bg_resume()
        Window.clearcolor = [0, 0, 0, 1]

    def stop(self):
        if self._foundation_process:
            self._foundation_process.terminate()
        super(OverlayWindowGuiMixin, self).stop()

    def gui_setup(self):
        super(OverlayWindowGuiMixin, self).gui_setup()
        
        if self.config.show_foundation and \
                self.config.foundation_image and \
                os.path.exists(self.config.foundation_image): 
            cmd = ['pngview', '-l', str(self.config.dispmanx_foundation_layer),
                   '-n', self.config.foundation_image]
            self._foundation_process = subprocess.Popen(cmd)

        self.overlay_mode = self.config.overlay_mode



import kivy.uix.boxlayout

from ebs.iot.linuxnode.tables.renderable import BasicRenderableTable
from ebs.iot.linuxnode.tables.spec import BasicTableSpec


class TranslatableTableSpec(BasicTableSpec):
    def __init__(self, *args, **kwargs):
        self._languages = kwargs.pop('languages', [])
        self._next_language = -1
        self._metadata = kwargs.pop('i18n_metadata', {})
        super(TranslatableTableSpec, self).__init__(*args, **kwargs)

    def install_metadata(self, metadata):
        self._metadata.update(metadata)

    def install(self):
        super(TranslatableTableSpec, self).install()
        self._i18n_install_languages()

    def _i18n_install_languages(self):
        for language in self._languages:
            self.parent.log.debug("Installing Language {0} for Table {1}".format(language, self.name))
            self.i18n_install_language(language)

    @property
    def i18n_metadata(self):
        if not self._metadata:
            self._metadata = {
                'project': None,
                'version': None,
                'msgid_bugs_address': None,
                'language_team': None,
                'last_translator': None,
                'copyright_holder': None,
            }
        return self._metadata

    def i18n_install_language(self, language):
        if language not in self._languages:
            self._languages.append(language)
        self.parent.node.i18n.install_context(self.name, language, metadata=self.i18n_metadata)

    @property
    def languages(self):
        return self._languages

    @property
    def next_language(self):
        self._next_language += 1
        if self._next_language == len(self._languages):
            self._next_language = 0
        return self._languages[self._next_language]

    def i18n_translator(self, language):
        return self.parent.node.i18n.translator(self.name, language)


class TranslatableRenderableTable(BasicRenderableTable):
    def __init__(self, *args, **kwargs):
        self._i18n_current = None
        self._i18n_language = None
        self._i18n_handlers = []
        super(TranslatableRenderableTable, self).__init__(*args, **kwargs)

    def install_language_change_handler(self, handler):
        self._i18n_handlers.append(handler)

    def install(self):
        super(TranslatableRenderableTable, self).install()
        self.language = self.spec.next_language
        self.install_language_change_handler(self._redraw_title)
        self.install_language_change_handler(self._redraw_header)

    @property
    def spec(self) -> TranslatableTableSpec:
        return self._spec

    @property
    def language(self):
        return self._i18n_language
    
    @language.setter
    def language(self, value):
        if value not in self.spec.languages:
            raise ValueError("Language '{0}' not installed!".format(value))
        self._i18n_language = value
        self._i18n_current = self.spec.i18n_translator(value)
        for handler in self._i18n_handlers:
            handler(self, value)

    def next_language(self):
        self.language = self.spec.next_language

    @property
    def i18n(self):
        return self._i18n_current

    def preprocess(self, value, spec=None):
        if spec and hasattr(spec, 'i18n') and not spec.i18n:
            return value
        else:
            return self.i18n(value)

    def _redraw_title(self, *_):
        pass

    def _redraw_header(self, *_):
        new_header = self._build_header(palette=self.palette)
        self._gui_table.remove_widget(self._gui_table.children[2])
        self._gui_table.add_widget(new_header, 2)





from twisted import logger

from .spec import BasicTableSpec


class BasicTableEntry(object):
    def __init__(self, data):
        self._data = data
        self._parent = None

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def data(self):
        return self._data


class BasicTable(object):
    def __init__(self, node, spec):
        self._node = node
        self._log = None
        self._spec = spec
        self._entries = []

    def install(self):
        self._spec.install()

    @property
    def node(self):
        return self._node

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="table.{0}".format(self._spec.name),
                                      source=self)
        return self._log

    @property
    def spec(self) -> BasicTableSpec:
        return self._spec

    def update(self, entries, incremental=False):
        if incremental:
            raise NotImplementedError

        self._entries = []
        enable_dedup = bool(self.spec.dedup_keys)
        dedup_record = []

        for entry in entries:
            if enable_dedup:
                tags = tuple([getattr(entry, x) for x in self.spec.dedup_keys])
                if tags in dedup_record:
                    continue
                dedup_record.append(tags)
            entry.parent = self
            self._entries.append(entry)
        self.log.info("Extracted {0} Entries from API response".format(len(self._entries)))



import os

import pqueue
from twisted.internet.defer import succeed


class ApiPersistentActionQueue(object):
    def __init__(self, api_engine, prefix=None):
        self._prefix = prefix
        self._api_engine = api_engine
        self._api_queue = None

    def process(self):
        while True:
            try:
                api_func_name, args = self.api_queue.get_nowait()
                self._api_engine.log.info(
                    "Executing persistent API action : {func_name}, {args}",
                    func_name=api_func_name, args=args
                )
                getattr(self._api_engine, api_func_name)(*args)
            except pqueue.Empty:
                break
        return succeed(True)

    def enqueue_action(self, api_func_name, *args):
        self._api_engine.log.info(
            "Enqueuing API action to disk : {func_name}, {args}",
            func_name=api_func_name, args=args
        )
        self.api_queue.put((api_func_name, args))

    @property
    def api_queue(self):
        if not self._api_queue:
            self._api_queue = pqueue.Queue(
                self._api_queue_dir,
                tempdir=os.path.join(self._api_queue_dir, 'tmp')
            )
        return self._api_queue

    @property
    def _api_queue_dir(self):
        dir_name = 'apiqueue'
        if self._prefix:
            dir_name = '-'.join([self._prefix, dir_name])
        _api_queue_dir = os.path.join(self._api_engine.cache_dir, dir_name)
        _api_queue_tmp_dir = os.path.join(_api_queue_dir, 'tmp')
        os.makedirs(_api_queue_tmp_dir, exist_ok=True)
        return _api_queue_dir





import math

from kivy.uix.widget import Widget
from kivy.uix.relativelayout import RelativeLayout

from kivy.properties import BooleanProperty
from kivy.properties import StringProperty
from kivy.animation import Animation

from .colors import ColorBoxLayout
from .image import StandardImage


class ImageGallery(ColorBoxLayout):
    visible = BooleanProperty(False)
    transition = StringProperty('in_out_elastic')

    def __init__(self, parent_layout=None, animation_vector=(0, 1),
                 exit_retrace=False, **kwargs):
        self.parent_layout = parent_layout
        self._animation_vector = animation_vector
        self._exit_retrace = exit_retrace

        self._animation_layer = None
        self._exit_animation = None
        self._entry_animation = None
        self._anim_distance_x = None
        self._anim_distance_y = None
        self._animation_distance = None

        super(ImageGallery, self).__init__(bgcolor=[0, 0, 0, 1], **kwargs)

        self.bind(visible=self._set_visibility)
        self.bind(size=self._calculate_animation_distance)
        self.bind(transition=self._reset_transitions)

        self._image = None

    def _set_visibility(self, *args):
        if self.visible:
            self.show()
        else:
            self.hide()

    def show(self):
        self.parent_layout.add_widget(self)

    def hide(self):
        self.clear_widgets()
        self._image = None
        self.animation_layer.clear_widgets()
        self._animation_layer = None
        self._reset_transitions()
        self.parent_layout.clear_widgets()

    def _reset_transitions(self, *args):
        self._exit_animation = None
        self._entry_animation = None

    def _calculate_animation_distance(self, *args):
        self._anim_distance_x = (self._animation_vector[0] * self.width)
        self._anim_distance_y = (self._animation_vector[1] * self.height)
        self._animation_distance = math.sqrt(self._anim_distance_x ** 2 + self._anim_distance_y ** 2)
        self._reset_transitions()

    @property
    def animation_distance(self):
        if not self._animation_distance:
            self._calculate_animation_distance()
        return self._animation_distance

    @property
    def animation_layer(self):
        if not self._animation_layer:
            self._animation_layer = RelativeLayout()
            self.parent_layout.add_widget(self.animation_layer,
                                          len(self.parent_layout.children) - 1)
        return self._animation_layer

    @property
    def exit_animation(self):
        if not self._exit_animation:
            _ = self.animation_distance

            def _when_done(_, instance):
                if not self._animation_layer:
                    return
                self.animation_layer.remove_widget(instance)
                if hasattr(instance, 'stop'):
                    instance.stop()

            if self._exit_retrace:
                sgn = -1
            else:
                sgn = 1
            self._exit_animation = Animation(y=self.pos[1] + sgn * self._anim_distance_y,
                                             x=self.pos[0] + sgn * self._anim_distance_x,
                                             t=self.transition, duration=2)
            self._exit_animation.bind(on_complete=_when_done)
        return self._exit_animation

    @property
    def entry_animation(self):
        if not self._entry_animation:
            _ = self.animation_distance

            def _when_done(_, instance):
                if not self._animation_layer:
                    return
                self.animation_layer.remove_widget(instance)
                instance.size_hint = (1, 1)
                self.add_widget(instance)
            self._entry_animation = Animation(y=self.pos[1], x=self.pos[0],
                                              t=self.transition, duration=2)
            self._entry_animation.bind(on_complete=_when_done)
        return self._entry_animation

    def _detach_image(self):
        if self._image:
            if self._image.parent == self:
                self.remove_widget(self._image)
            elif self._image.parent == self.animation_layer:
                self.animation_layer.remove_widget(self._image)

    @property
    def current(self):
        return self._image

    @current.setter
    def current(self, value):
        if value is None:
            self.visible = False
            self._detach_image()
            self._image = None
            return

        if self._image:
            pos = self._image.pos
            self._detach_image()
            self._image.size_hint = (None, None)
            self._image.pos = pos
            self.animation_layer.add_widget(self._image)
            self.exit_animation.start(self._image)

        if isinstance(value, Widget):
            self._image = value
        else:
            self._image = StandardImage(source=value, allow_stretch=True,
                                        keep_ratio=True, anim_delay=0.08)

        if not self.visible:
            self.add_widget(self._image)
            self.visible = True
            return

        self._image.size_hint = (None, None)
        self._image.size = self.size
        self._image.pos = (self.pos[0] - self._anim_distance_x,
                           self.pos[1] - self._anim_distance_y)
        self.animation_layer.add_widget(self._image)
        self.entry_animation.start(self._image)
        self.clear_widgets()




from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.stacklayout import StackLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout


class BaseGuiStructureMixin(object):
    def __init__(self, *args, **kwargs):
        self._gui_root = None
        self._gui_structure_root = None
        self._gui_primary_root = None
        self._gui_footer = None
        self._gui_anchor_br = None
        self._gui_anchor_bl = None
        self._gui_anchor_tr = None
        self._gui_status_stack = None
        self._gui_notification_stack = None
        self._gui_notification_row = None
        self._gui_debug_stack = None
        self._gui_content_root = None
        self._gui_main_content = None
        self._gui_sidebar_right = None
        self._gui_sidebar_right_users = set()
        self._gui_sidebar_left = None
        self._gui_animation_layer = None
        super(BaseGuiStructureMixin, self).__init__(*args, **kwargs)

    @property
    def gui_anchor_bottom_right(self):
        if not self._gui_anchor_br:
            self._gui_anchor_br = AnchorLayout(anchor_x='right',
                                               anchor_y='bottom',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_br)
        return self._gui_anchor_br

    @property
    def gui_anchor_bottom_left(self):
        if not self._gui_anchor_bl:
            self._gui_anchor_bl = AnchorLayout(anchor_x='left',
                                               anchor_y='bottom',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_bl)
        return self._gui_anchor_bl

    @property
    def gui_anchor_top_right(self):
        if not self._gui_anchor_tr:
            self._gui_anchor_tr = AnchorLayout(anchor_x='right',
                                               anchor_y='top',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_tr)
        return self._gui_anchor_tr

    @property
    def gui_status_stack(self):
        if not self._gui_status_stack:
            self._gui_status_stack = StackLayout(orientation='bt-rl',
                                                 padding='8sp')
            self.gui_anchor_bottom_right.add_widget(self._gui_status_stack)
        return self._gui_status_stack

    @property
    def gui_notification_stack(self):
        if not self._gui_notification_stack:
            self._gui_notification_stack = GridLayout(cols=1,
                                                      padding='8sp',
                                                      spacing='8sp',
                                                      size_hint_y=None,)

            def _set_height(_, mheight):
                self.gui_notification_stack.height = mheight
            self.gui_notification_stack.bind(minimum_height=_set_height)
            self.gui_anchor_bottom_left.add_widget(self._gui_notification_stack)
        return self._gui_notification_stack

    @property
    def gui_notification_row(self):
        if not self._gui_notification_row:
            self._gui_notification_row = StackLayout(orientation='lr-bt',
                                                     spacing='8sp')
            self.gui_notification_stack.add_widget(self._gui_notification_row)
        return self._gui_notification_row

    def gui_notification_update(self):
        self.gui_notification_row.do_layout()
        self.gui_notification_stack.do_layout()

    @property
    def gui_debug_stack(self):
        if not self._gui_debug_stack:
            self._gui_debug_stack = StackLayout(orientation='tb-rl',
                                                padding='8sp')
            self.gui_anchor_top_right.add_widget(self._gui_debug_stack)
        return self._gui_debug_stack

    @property
    def gui_footer(self):
        if not self._gui_footer:
            _ = self.gui_primary_root
            self._gui_footer = BoxLayout(
                orientation='vertical', size_hint=(1, None),
                height=80, padding=['0sp', '0sp', '0sp', '8sp']
            )
        return self._gui_footer

    def gui_footer_show(self):
        if not self._gui_footer.parent:
            self.gui_structure_root.add_widget(self._gui_footer)

    def gui_footer_hide(self):
        if self._gui_footer.parent:
            self.gui_structure_root.remove_widget(self._gui_footer)

    @property
    def gui_primary_root(self):
        if not self._gui_primary_root:
            self._gui_primary_root = FloatLayout()
            self.gui_structure_root.add_widget(self._gui_primary_root)
        return self._gui_primary_root

    @property
    def gui_structure_root(self):
        if not self._gui_structure_root:
            self._gui_structure_root = BoxLayout(orientation='vertical')
            self.gui_root.add_widget(self._gui_structure_root)
        return self._gui_structure_root

    @property
    def gui_main_content(self):
        if not self._gui_main_content:
            self._gui_main_content = FloatLayout()
            self.gui_content_root.add_widget(self._gui_main_content)
        return self._gui_main_content

    @property
    def gui_sidebar_right(self):
        if not self._gui_sidebar_right:
            x_hint = self.config.sidebar_width / (1 - self.config.sidebar_width)
            self._gui_sidebar_right = BoxLayout(orientation='vertical',
                                                size_hint=(x_hint, 1))
        return self._gui_sidebar_right

    def gui_sidebar_right_show(self, key):
        if not key:
            key = 'unspecified'
        self._gui_sidebar_right_users.add(key)
        if not self.gui_sidebar_right.parent:
            self.log.debug("Showing right sidebar")
            self.gui_content_root.add_widget(self.gui_sidebar_right)

    def gui_sidebar_right_hide(self, key):
        if key not in self._gui_sidebar_right_users:
            return
        self._gui_sidebar_right_users.remove(key)
        if len(self._gui_sidebar_right_users):
            return
        if self.gui_sidebar_right.parent:
            self.log.debug("Hiding right sidebar")
            self.gui_content_root.remove_widget(self.gui_sidebar_right)

    # @property
    # def gui_sidebar_left(self):
    #     if not self._gui_sidebar_left:
    #         self._gui_sidebar_left = FloatLayout(size_hint=((0.28/0.72), 1))
    #     return self._gui_sidebar_left
    #
    # def gui_sidebar_left_show(self):
    #     if not self.gui_sidebar_left.parent:
    #         print(self.gui_content_root.children)
    #         self.gui_content_root.add_widget(
    #             self.gui_sidebar_left,
    #             index=len(self.gui_content_root.children)
    #         )
    #         self.gui_content_root.do_layout()
    #         print(self.gui_content_root.children)
    #
    # def gui_sidebar_left_hide(self):
    #     if self.gui_sidebar_left.parent:
    #         self.gui_content_root.remove_widget(self.gui_sidebar_left)

    @property
    def gui_content_root(self):
        if not self._gui_content_root:
            self._gui_content_root = GridLayout(rows=1, spacing=10)
            self.gui_root.add_widget(self._gui_content_root,
                                     len(self.gui_root.children) - 1)
        return self._gui_content_root

    @property
    def gui_animation_layer(self):
        if not self._gui_animation_layer:
            self._gui_animation_layer = FloatLayout()
            self.gui_root.add_widget(self._gui_animation_layer,
                                     len(self.gui_root.children) - 1)
        return self._gui_animation_layer

    @property
    def gui_root(self):
        if not self._gui_root:
            self._gui_root = FloatLayout()
        return self._gui_root






import objgraph


previous_snapshot = {}


def take_snapshot():
    global previous_snapshot
    snapshot = {t: c for (t, c) in objgraph.most_common_types(limit=100)}
    print("Took snapshot of objgraph")
    if previous_snapshot:
        delta = {t: snapshot[t] - previous_snapshot.get(t, 0) for t in snapshot.keys()}
        print("Types with the most changes : ")
        for t in sorted(delta.keys(), key=lambda x: delta[x], reverse=True):
            if delta[t]:
                print("{0:20} {1}".format(t, delta[t]))
    previous_snapshot = snapshot


def show_chain(ob):
    objgraph.show_chain(objgraph.find_backref_chain(ob, objgraph.is_proper_module))





# foundation    ()
# backdrop      1
# background    2
# backdrop      3
# video         4
# app           5


import os
import pkg_resources
from six.moves.configparser import ConfigParser
from appdirs import user_config_dir


class IoTNodeConfig(object):
    _appname = 'iotnode'
    _root = os.path.abspath(os.path.dirname(__file__))
    _roots = [_root]
    _config_file = os.path.join(user_config_dir(_appname), 'config.ini')
    _sys_config_file = os.path.join('/etc/raspap/custom.ini')

    def __init__(self):
        self._config = ConfigParser()
        print("Reading Config File {}".format(self._config_file))
        self._config.read(self._config_file)
        self._sys_config = ConfigParser()
        print("Reading System Config File {}".format(self._sys_config_file))
        self._sys_config.read(self._sys_config_file)
        self._config_apply_init()
        print("EBS IOT Linux Node, version {0}".format(self.linuxnode_version))

    @property
    def linuxnode_version(self):
        return pkg_resources.get_distribution('ebs-iot-linuxnode').version

    def _config_apply_init(self):
        self._apply_display_layer()

    def _write_config(self):
        with open(self._config_file, 'w') as configfile:
            self._config.write(configfile)

    def _check_section(self, section):
        if not self._config.has_section(section):
            self._config.add_section(section)
            self._write_config()

    # Paths
    @property
    def app_root(self):
        return self._root

    @property
    def app_resources(self):
        return os.path.join(self.app_root, 'resources')

    @property
    def roots(self):
        return self._roots

    def get_path(self, filepath):
        for root in self._roots:
            if os.path.exists(os.path.join(root, filepath)):
                return os.path.join(root, filepath)
        return filepath

    # Platform
    @property
    def platform(self):
        return self._config.get('platform', 'platform', fallback=None)

    # Debug
    @property
    def debug(self):
        return self._config.getboolean('debug', 'debug', fallback=False)

    @property
    def gui_log_display(self):
        return self._config.getboolean('debug', 'gui_log_display', fallback=False)

    @property
    def gui_log_level(self):
        return self._config.get('debug', 'gui_log_level', fallback='info')

    # Display
    @property
    def fullscreen(self):
        return self._config.getboolean('display', 'fullscreen', fallback=True)

    @property
    def overlay_mode(self):
        return self._config.getboolean('display', 'overlay_mode', fallback=False)

    @property
    def sidebar_width(self):
        return self._config.getfloat('display', 'sidebar_width', fallback=0.3)

    @property
    def show_foundation(self):
        return self._config.getboolean('display-rpi', 'show_foundation', fallback=True)

    @property
    def dispmanx_foundation_layer(self):
        return self._config.getint('display-rpi', 'dispmanx_foundation_layer', fallback=1)

    @property
    def foundation_image(self):
        return self._config.get('display-rpi', 'foundation_image', fallback=None)

    @property
    def background(self):
        return self._config.get('display', 'background', fallback='images/background.png')

    @background.setter
    def background(self, value):
        self._check_section('display')
        self._config.set('display', 'background', value)
        self._write_config()

    @property
    def background_external_player(self):
        if self.platform == 'rpi':
            return self._config.getboolean('display-rpi', 'background_external_player', fallback=False)

    @property
    def background_dispmanx_layer(self):
        return self._config.getint('display-rpi', 'background_dispmanx_layer', fallback=2)

    @property
    def app_dispmanx_layer(self):
        if self.platform != 'rpi':
            raise AttributeError("dispmanx layer is an RPI thing")
        return self._config.getint('display-rpi', 'dispmanx_app_layer', fallback=5)

    def _apply_display_layer(self):
        if self.platform == 'rpi':
            os.environ.setdefault('KIVY_BCM_DISPMANX_LAYER', str(self.app_dispmanx_layer))

    # ID
    @property
    def node_id_getter(self):
        return self._config.get('id', 'getter', fallback='uuid')

    @property
    def node_id_interface(self):
        return self._config.get('id', 'interface', fallback=None)

    @property
    def node_id_override(self):
        return self._config.get('id', 'override', fallback=None)

    @property
    def node_id_display(self):
        return self._config.getboolean('id', 'display', fallback=False)

    @property
    def node_id_display_frequency(self):
        return self._config.getint('id', 'display_frequency', fallback=0)

    @property
    def node_id_display_duration(self):
        return self._config.getint('id', 'display_duration', fallback=15)

    # HTTP
    @property
    def http_max_concurrent_requests(self):
        return self._config.getint('http', 'max_concurrent_requests', fallback=1)

    @property
    def http_max_background_downloads(self):
        return self._config.getint('http', 'max_background_downloads', fallback=1)

    @property
    def http_max_concurrent_downloads(self):
        return self._config.getint('http', 'max_concurrent_downloads', fallback=1)

    @property
    def http_proxy_host(self):
        return self._sys_config.get('NetworkProxyConfiguration', 'host', fallback=None)

    @property
    def http_proxy_port(self):
        return self._sys_config.getint('NetworkProxyConfiguration', 'port', fallback=0)

    @property
    def http_proxy_user(self):
        return self._sys_config.get('NetworkProxyConfiguration', 'user', fallback=None)

    @property
    def http_proxy_pass(self):
        return self._sys_config.get('NetworkProxyConfiguration', 'pass', fallback=None)

    @property
    def http_proxy_enabled(self):
        return self.http_proxy_host is not None

    @property
    def http_proxy_auth(self):
        if not self.http_proxy_user:
            return None
        if not self.http_proxy_pass:
            return self.http_proxy_user
        return "{0}:{1}".format(self.http_proxy_user, self.http_proxy_pass)

    @property
    def http_proxy_url(self):
        url = self.http_proxy_host
        if self.http_proxy_port:
            url = "{0}:{1}".format(url, self.http_proxy_port)
        if self.http_proxy_auth:
            url = "{0}@{1}".format(self.http_proxy_auth, url)
        return url

    # Resource Manager
    @property
    def resource_prefetch_retries(self):
        return self._config.getint('resources', 'prefetch_retries', fallback=3)
    
    @property
    def resource_prefetch_retry_delay(self):
        return self._config.getint('resources', 'prefetch_retry_delay', fallback=5)

    # Cache
    @property
    def cache_max_size(self):
        return self._config.getint('cache', 'max_size', fallback='10000000')

    # Video
    @property
    def video_external_player(self):
        if self.platform == 'rpi':
            return self._config.getboolean('video-rpi', 'external_player', fallback=False)

    @property
    def video_dispmanx_layer(self):
        if self.platform == 'rpi':
            return self._config.getint('video-rpi', 'dispmanx_video_layer', fallback=4)

    @property
    def video_show_backdrop(self):
        if self.platform == 'rpi':
            return self._config.getboolean('video-rpi', 'show_backdrop', fallback=False)

    @property
    def video_backdrop_dispmanx_layer(self):
        if self.platform == 'rpi':
            return self._config.getint('video-rpi', 'dispmanx_video_layer', fallback=1)

    # Browser
    @property
    def browser_show_default(self):
        return self._config.getboolean('browser', 'show_default', fallback=False)

    @browser_show_default.setter
    def browser_show_default(self, value):
        self._check_section('browser')
        if value:
            value = 'yes'
        else:
            value = 'no'
        self._config.set('browser', 'show_default', value)
        self._write_config()

    @property
    def browser_default_url(self):
        return self._config.get('browser', 'default_url', fallback='http://www.google.com')

    @browser_default_url.setter
    def browser_default_url(self, value):
        self._check_section('browser')
        self._config.set('browser', 'default_url', value)
        self._write_config()

    # Fonts
    @property
    def text_font_name(self):
        return self.get_path(self._config.get('text', 'font_name', fallback=None)
)
    @property
    def text_use_fcm(self):
        return self._config.getboolean('text', 'use_fcm', fallback=False)
    
    @property
    def text_fcm_system(self):
        return self._config.getboolean('text', 'fcm_system', fallback=True)
    
    @property
    def text_fcm_fonts(self):
        return self.get_path(self._config.get('text', 'fcm_fonts', fallback=None))


class ConfigMixin(object):
    def __init__(self, *args, **kwargs):
        global current_config
        self._config = current_config
        super(ConfigMixin, self).__init__(*args, **kwargs)

    @property
    def config(self) -> IoTNodeConfig:
        return self._config





from twisted.internet.defer import inlineCallbacks
import treq


class StandaloneActual(object):
    @inlineCallbacks
    def http_get(self, url, **kwargs):
        d = yield treq.request('get', url, **kwargs)
        return d

    @inlineCallbacks
    def http_post(self, url, **kwargs):
        response = yield treq.post(url, **kwargs)
        data = yield response.json()
        return data


class StandaloneConfig(object):
    def __init__(self, *args, **kwargs):
        pass





import os
from kivy.uix.video import Video
from twisted.internet.defer import Deferred

from .widgets.image import StandardImage
from .widgets.colors import ColorBoxLayout
from .widgets.pdfplayer import PDFPlayer

from .log import NodeLoggingMixin
from .background import OverlayWindowGuiMixin
from .externalplayer import ExternalMediaPlayer
from .externalplayer import BackdropManager


class MediaPlayerBusy(Exception):
    def __init__(self, now_playing, collision_count):
        self.now_playing = now_playing
        self.collision_count = collision_count

    def __repr__(self):
        return "<MediaPlayerBusy Now Playing {0}" \
               "".format(self.now_playing)


class MediaPlayerMixin(NodeLoggingMixin):
    _media_extentions_image = ['.png', '.jpg', '.bmp', '.gif', '.jpeg']
    _media_extentions_video = []

    def __init__(self, *args, **kwargs):
        super(MediaPlayerMixin, self).__init__(*args, **kwargs)
        self._media_player_deferred = None
        self._mediaplayer_now_playing = None
        self._end_call = None
        self._mediaplayer_collision_count = 0

    def media_play(self, content, duration=None, loop=False, interval=None):
        # Play the media file at filepath. If loop is true, restart the media
        # when it's done. You probably would want to provide a duration with
        # an image or with a looping video, not otherwise.
        if self._mediaplayer_now_playing:
            self._mediaplayer_collision_count += 1
            if self._mediaplayer_collision_count > 30:
                self.media_stop(forced=True)
            raise MediaPlayerBusy(self._mediaplayer_now_playing,
                                  self._mediaplayer_collision_count)
        self._mediaplayer_collision_count = 0
        if hasattr(content, 'filepath'):
            content = content.filepath
        if not os.path.exists(content):
            self.log.warn("Could not find media to play at {filepath}",
                          filepath=content)
            return
        if duration:
            self._end_call = self.reactor.callLater(duration, self.media_stop)
        self._mediaplayer_now_playing = os.path.basename(content)
        self.gui_bg_pause()
        if os.path.splitext(content)[1] in self._media_extentions_image:
            self.log.debug("Showing image {filename}",
                           filename=os.path.basename(content))
            self._media_play_image(content)
        elif os.path.splitext(content)[1] in ['.pdf']:
            self.log.debug("Showing pdf {filename}",
                           filename=os.path.basename(content))
            self._media_play_pdf(content, interval=interval)
        else:
            self.log.debug("Starting video {filename}",
                           filename=os.path.basename(content))
            self._media_play_video(content, loop)
        self._media_player_deferred = Deferred()
        return self._media_player_deferred

    def _media_play_image(self, filepath):
        raise NotImplementedError

    def _media_play_pdf(self, filepath, interval=None):
        raise NotImplementedError

    def _media_play_video(self, filepath, loop=False):
        raise NotImplementedError

    def media_stop(self, forced=False):
        self.log.info("End Offset by {0} collisions."
                      "".format(self._mediaplayer_collision_count))
        self._mediaplayer_collision_count = 0

        def _resume_bg():
            if not self._mediaplayer_now_playing:
                self.gui_bg_resume()
                self.gui_mediaview.make_transparent()
        self.reactor.callLater(1.5, _resume_bg)

        if self._end_call and self._end_call.active():
            self._end_call.cancel()

        if self._mediaplayer_now_playing:
            self._mediaplayer_now_playing = None

        if self._media_player_deferred:
            self._media_player_deferred.callback(forced)
            self._media_player_deferred = None

    def stop(self):
        self.media_stop(forced=True)
        super(MediaPlayerMixin, self).stop()


class MediaPlayerGuiMixin(OverlayWindowGuiMixin):
    def __init__(self, *args, **kwargs):
        super(MediaPlayerGuiMixin, self).__init__(*args, **kwargs)
        self._media_player_external = None
        self._media_player_backdrop = BackdropManager()
        self._media_playing = None
        self._gui_mediaview = None

    def _media_play_image(self, filepath):
        self._media_playing = StandardImage(source=filepath,
                                            allow_stretch=True,
                                            keep_ratio=True)
        self.gui_mediaview.add_widget(self._media_playing)

    def _media_play_pdf(self, filepath, interval=None):
        self._media_playing = PDFPlayer(source=filepath,
                                        temp_dir=self.temp_dir)
        if interval:
            self._media_playing.interval = interval
        self.gui_mediaview.add_widget(self._media_playing)

    def _media_play_video(self, *args, **kwargs):
        if self.config.video_external_player:
            self._media_play_video_omxplayer(*args, **kwargs)
        else:
            self._media_play_video_native(*args, **kwargs)

    def _media_play_video_native(self, filepath, loop=False):
        if loop:
            eos = 'loop'
        else:
            eos = 'stop'
        self._media_playing = Video(source=filepath, state='play',
                                    eos=eos, allow_stretch=True)
        self._media_playing.opacity = 0
        self.gui_mediaview.make_opaque()

        def _while_playing(*_):
            self._media_playing.opacity = 1
        self._media_playing.bind(texture=_while_playing)

        def _when_done(*_):
            self.media_stop()
        self._media_playing.bind(eos=_when_done)

        self.gui_mediaview.add_widget(self._media_playing)

    def _media_play_video_omxplayer(self, filepath, loop=False):
        self._media_playing = ExternalMediaPlayer(
            filepath,
            (self.gui_mediaview.x, self.gui_mediaview.y,
             self.gui_mediaview.width, self.gui_mediaview.height),
            self.media_stop, self, layer=None, loop=False,
            dbus_name='org.mpris.MediaPlayer2.omxplayer2'
        )

    def media_stop(self, forced=False):
        print("Stopping Media : {0}".format(self._media_playing))
        if isinstance(self._media_playing, Video):
            self._media_playing.unload()
        elif isinstance(self._media_playing, ExternalMediaPlayer):
            self._media_playing.force_stop()
            self._media_playing = None
        elif isinstance(self._media_playing, StandardImage):
            pass
        elif isinstance(self._media_playing, PDFPlayer):
            self._media_playing.stop()
        self.gui_mediaview.clear_widgets()
        MediaPlayerMixin.media_stop(self, forced=forced)

    def stop(self):
        if self._media_player_backdrop:
            self._media_player_backdrop.close()
        super(MediaPlayerGuiMixin, self).stop()

    @property
    def gui_mediaview(self):
        if self._gui_mediaview is None:
            self._gui_mediaview = ColorBoxLayout(bgcolor=(0, 0, 0, 0))
            self.gui_main_content.add_widget(self._gui_mediaview)
            
            if self.config.video_show_backdrop:
                self._media_player_backdrop.start(
                    self.config.video_backdrop_dispmanx_layer,
                    self.gui_mediaview.x, self.gui_mediaview.y,
                    self.gui_mediaview.width, self.gui_mediaview.height
                )

                def _backdrop_geometry(widget, _):
                    self._media_player_backdrop.set_geometry(
                        widget.x, widget.y, widget.width, widget.height
                    )
                self.gui_mediaview.bind(size=_backdrop_geometry,
                                        pos=_backdrop_geometry)

            def _child_geometry(widget, _):
                if isinstance(self._media_playing, ExternalMediaPlayer):
                    self._media_playing.set_geometry(
                        widget.x, widget.y, widget.width, widget.height
                    )
            self.gui_mediaview.bind(size=_child_geometry,
                                    pos=_child_geometry)
        return self._gui_mediaview

    def gui_setup(self):
        super(MediaPlayerGuiMixin, self).gui_setup()
        _ = self.gui_mediaview





from colorthief import ColorThief
from PIL import Image as PILImage

from kivy.uix.image import Image

from .colors import BackgroundColorMixin

from kivy.graphics.opengl import glGetIntegerv
from kivy.graphics.opengl import GL_MAX_TEXTURE_SIZE
_image_max_size = glGetIntegerv(GL_MAX_TEXTURE_SIZE)[0]


class SizeProofImage(Image):
    def __init__(self, **kwargs):
        source = kwargs.get('source', None)
        if source:
            PILImage.MAX_IMAGE_PIXELS = None
            im = PILImage.open(source)
            size = im.size
            sf = max([float(s) / _image_max_size for s in size])
            if sf > 1:
                target = [int(s / sf) for s in size]
                print("Resizing image {1} to {2} {0}"
                      "".format(source, size, target))
                im = im.resize(target, PILImage.ANTIALIAS)
                im.save(source)
            im.close()
            del im
        Image.__init__(self, **kwargs)


StandardImage = SizeProofImage


class BleedImage(BackgroundColorMixin, StandardImage):
    def __init__(self, **kwargs):
        bgcolor = kwargs.pop('bgcolor', 'auto')
        StandardImage.__init__(self, **kwargs)
        BackgroundColorMixin.__init__(self)
        if bgcolor == 'auto':
            self._autoset_bg_color()
            self.bind(source=self._autoset_bg_color)
        else:
            self.bgcolor = bgcolor

    def _autoset_bg_color(self, *_):
        color = ColorThief(self.source).get_color(5)
        self.bgcolor = (x/255 for x in color)





import os
import io
import sys
import time
from twisted import logger
from twisted.logger import formatEvent
from twisted.logger import LogLevel
from twisted.logger import LogLevelFilterPredicate
from twisted.logger import FilteringLogObserver
from twisted.logger import textFileLogObserver
from twisted.logger import STDLibLogObserver

from appdirs import user_log_dir
from datetime import datetime
from collections import deque
from kivy.core.window import Window
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.effects.scroll import ScrollEffect
from kivy.utils import get_hex_from_color
from .widgets.labels import ColorLabel

from .config import ConfigMixin
from .basemixin import BaseMixin
from .basemixin import BaseGuiMixin

import logging
logging.basicConfig(level=logging.DEBUG)


class NodeLoggingMixin(ConfigMixin, BaseMixin):
    _log = None

    def __init__(self, *args, **kwargs):
        super(NodeLoggingMixin, self).__init__(*args, **kwargs)
        self._log_file = None
        self.log_prune()
        self._log = logger.Logger(namespace=self._appname,
                                  source=self)
        self.reactor.callWhenRunning(self._start_logging)

    def _observers(self):
        if self.config.debug:
            level = LogLevel.debug
        else:
            level = LogLevel.info

        return [
            # STDLibLogObserver(),
            FilteringLogObserver(
                textFileLogObserver(sys.stdout),
                predicates=[LogLevelFilterPredicate(LogLevel.warn)]
            ),
            FilteringLogObserver(
                textFileLogObserver(io.open(self.log_file, 'a')),
                predicates=[LogLevelFilterPredicate(level)]
            ),

        ]

    def _start_logging(self):
        # TODO Mention that docs don't say reactor should be running
        # TODO Mention that docs are confusing about how extract works
        # TODO Find out about a functional print to console observer
        # TODO Mention problem with IOBase vs TextIOWrapper
        # TODO log_source is not set when logger instantiated in __init__
        logger.globalLogBeginner.beginLoggingTo(self._observers())
        self.log.info("Logging to {logfile}", logfile=self.log_file)

    @property
    def log(self):
        return self._log

    @property
    def log_file(self):
        if not self._log_file:
            self._log_file = os.path.join(
                self.log_dir, 
                'runlog_{0}'.format(datetime.today().strftime('%d%m%y'))
            )
        return self._log_file

    def log_prune(self):
        for fname in self.log_files:
            fpath = os.path.join(self.log_dir, fname)
            mtime = os.path.getmtime(fpath)
            if time.time() - mtime > (7 * 24 * 60 * 60):
                os.remove(fpath)

    @property
    def log_files(self):
        return self._log_files()

    def _log_files(self):
        for filename in os.listdir(self.log_dir):
            if os.path.isfile(os.path.join(self.log_dir, filename)):
                yield filename

    @property
    def log_dir(self):
        os.makedirs(user_log_dir(self._appname), exist_ok=True)
        return user_log_dir(self._appname)


class LoggingGuiMixin(ConfigMixin, BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._gui_log = None
        self._gui_log_end = None
        self._gui_log_layout = None
        self._gui_log_scroll = None
        self._gui_log_lines = deque([], maxlen=100)
        super(LoggingGuiMixin, self).__init__(*args, **kwargs)

    def _observers(self):
        rv = NodeLoggingMixin._observers(self)
        if self.config.gui_log_display:
            rv.extend([self.gui_log_observer])
        return rv

    _level_ignore_map = {
        'trace': [],
        'debug': ['trace'],
        'info': ['trace', 'debug'],
        'warning': ['trace', 'debug', 'info'],
        'error': ['trace', 'debug', 'info', 'error'],
    }

    def gui_log_observer(self, event):
        ll = event['log_level'].name
        if ll in self._level_ignore_map[self.config.gui_log_level]:
            return
        msg = "[font=RobotoMono-Regular][{0:^8}][/font] {1} {2}".format(
            ll.upper(),
            datetime.fromtimestamp(event['log_time']).strftime("%d%m %H:%M:%S"),
            formatEvent(event)
        )
        color = None
        if ll == 'warn':
            color = [1, 1, 0, 1]
        elif ll == 'error':
            color = [1, 0, 0, 1]
        elif ll == 'critical':
            color = [1, 0, 0, 1]
        if color:
            color = get_hex_from_color(color)
            msg = '[color={0}]{1}[/color]'.format(color, msg)

        self._gui_log_lines.append(msg)
        self._gui_log.text = '\n'.join(self._gui_log_lines)

    @property
    def gui_log(self):
        if not self._gui_log:
            self._gui_log_scroll = ScrollView(
                size_hint=(None, None), bar_pos_y='right',
                bar_color=[1, 1, 1, 1], effect_cls=ScrollEffect,
                do_scroll_x=False, do_scroll_y=True)

            def _set_log_size(_, size):
                width = min(max(700, size[0] * 0.3), size[0])
                height = size[1] * 0.6
                self._gui_log_scroll.size = width, height
            self.gui_root.bind(size=_set_log_size)

            self._gui_log_layout = BoxLayout(orientation='vertical',
                                             size_hint=(1, None))

            self._gui_log = ColorLabel(
                size_hint=(1, 1), padding=(8, 8), bgcolor=[0, 0, 0, 0.2],
                halign='left', valign='top', markup=True, font_size='12sp',
            )

            self._gui_log_end = Label(size_hint=(1, None), height=1)

            def _set_label_height(_, size):
                if size[1] > Window.height * 0.3:
                    self._gui_log.height = size[1]
                else:
                    self._gui_log.height = Window.height * 0.3
            self._gui_log.bind(texture_size=_set_label_height)

            def _set_layout_height(_, size):
                if size[1] > Window.height * 0.3:
                    self._gui_log_layout.height = size[1]
                else:
                    self._gui_log_layout.height = Window.height * 0.3
                self._gui_log_scroll.scroll_to(self._gui_log_end)
            self._gui_log.bind(texture_size=_set_layout_height)

            def _set_text_width(_, size):
                self._gui_log.text_size = size[0], None
            self._gui_log.bind(size=_set_text_width)

            self._gui_log_layout.add_widget(self._gui_log)
            self._gui_log_layout.add_widget(self._gui_log_end)
            self._gui_log_scroll.add_widget(self._gui_log_layout)
            self.gui_debug_stack.add_widget(self._gui_log_scroll)
        return self._gui_log

    def gui_setup(self):
        super(LoggingGuiMixin, self).gui_setup()
        if not self.config.gui_log_display:
            return
        _ = self.gui_log





import subprocess

from omxplayer.player import OMXPlayer
from dbus.exceptions import DBusException


class BackdropManager(object):
    def __init__(self):
        self._process = None

    def start(self, layer=None, x=None, y=None, width=None, height=None):
        if not x:
            x = 10
        if not y:
            y = 10
        if not width:
            width = 1
        if not height:
            height = 1

        cmd = ['backdrop']
        if layer:
            cmd.extend(['-l', str(layer)])
        if x:
            cmd.extend(['-x', str(int(x))])
        if y:
            cmd.extend(['-y', str(int(y))])
        if width:
            cmd.extend(['-w', str(int(width))])
        if height:
            cmd.extend(['-h', str(int(height))])
        self._process = subprocess.Popen(cmd, stdin=subprocess.PIPE)

    def set_geometry(self, x, y, width, height):
        if not self._process:
            self.start(x, y, width, height)
        geometry = "{0},{1},{2},{3}\n".format(int(x), int(y),
                                              int(width), int(height))
        self._process.stdin.write(geometry.encode())
        self._process.stdin.flush()

    def close(self):
        if self._process:
            self._process.stdin.write('0,0,0,0'.encode())
            self._process.stdin.flush()
            self._process.terminate()


class ExternalMediaPlayer(object):
    def __init__(self, filepath, geometry, when_done, node,
                 layer=None, loop=False, dbus_name=None):
        self._player = None
        self._pposition = None
        self._pstate = None
        self._paused = False
        self._filepath = filepath
        self._cover = None
        self._node = node
        self._loop = loop
        self._dbus_name = dbus_name
        self._geometry = geometry
        self._when_done = when_done

        if not layer:
            layer = self._node.config.video_dispmanx_layer
        self._layer = layer

        self._launch_player()

    def _exit_handler(self, player, exit_state):
        if self._when_done and not self._paused:
            self._node.reactor.callFromThread(self._when_done)

    def _launch_player(self, paused=False):
        x, y, width, height = self._geometry
        args = [
            '--no-osd', '--aspect-mode', 'letterbox',
            '--layer', str(self._layer),
            '--win', '{0},{1},{2},{3}'.format(x, y, x + width, y + height),
            '--adev', 'hdmi',
        ]
        if self._loop:
            args.append('--loop')

        try:
            # print("Launching player on {0} with {1}".format(self._dbus_name, self._filepath))
            self._player = OMXPlayer(self._filepath, args=args, dbus_name=self._dbus_name)
            if paused:
                self._player.pause()
            # print("Installing exit handler")
            self._player.exitEvent = self._exit_handler
        except SystemError as e:
            self._player = None
            self._exit_handler(None, 1)
        # print("Done launching player")

    def force_stop(self):
        if self._player:
            # print("Force stopping player")
            self._player.quit()
            self._player = None

    def pause(self):
        if self._player:
            # print("Pausing player")
            self._pposition = self._player.position()
            self._pstate = self._player.playback_status()
            self._player.quit()
            self._player = None
            # print("Done pausing")

    def resume(self):
        if self._player or not self._pstate:
            return
        # print("Resuming player")
        self._launch_player(paused=True)
        if self._pposition:
            self._player.set_position(self._pposition)
        if self._pstate == "Playing":
            self._player.play()
        self._pstate = None
        # print("Done resume")

    def set_geometry(self, x, y, width, height):
        self._geometry = (x, y, width, height)
        if self._player:
            try:
                self._player.set_video_pos(x, y, x + width, y + height)
            except DBusException:
                pass





class BasicTablePalette(object):
    def __init__(self, **kwargs):
        self._color_cell_background = kwargs.pop('color_cell_background', None)
        self._color_cell_foreground = kwargs.pop('color_cell_foreground', None)
        self._color_header_cell_background = kwargs.pop('color_header_cell_background', None)
        self._color_header_cell_foreground = kwargs.pop('color_header_cell_foreground', None)
        self._color_grid_background = kwargs.pop('color_grid_background', None)

    @property
    def cell_background(self):
        return self._color_cell_background

    @property
    def cell_foreground(self):
        return self._color_cell_foreground

    @property
    def header_cell_background(self):
        return self._color_header_cell_background

    @property
    def header_cell_foreground(self):
        return self._color_header_cell_foreground

    @property
    def grid_background(self):
        return self._color_grid_background

    def __repr__(self):
        rv = "<{0}>".format(self.__class__.__name__)
        rv += "\n - grid_background : {}".format(self.grid_background)
        rv += "\n - cell_background : {}".format(self.cell_background)
        rv += "\n - cell_foreground : {}".format(self.cell_foreground)
        rv += "\n - header_cell_background : {}".format(self.header_cell_background)
        rv += "\n - header_cell_foreground : {}".format(self.header_cell_foreground)
        return rv


class BasicColumnSpec(object):
    def __init__(self, title, accessor, halign=None, font_bold=None,
                 width=None, width_hint=None, markup=False, i18n=True):
        self._parent = None
        self._title = title
        self._accessor = accessor
        self._halign = halign
        self._font_bold = font_bold
        self._width = width
        self._width_hint = width_hint
        self._markup = markup
        self._i18n = i18n

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def title(self):
        return self._title

    @property
    def accessor(self):
        return self._accessor

    @property
    def halign(self):
        return self._halign or 'center'

    @property
    def font_bold(self):
        if self._font_bold is None:
            return self.parent.font_bold
        return self._font_bold

    @property
    def width(self):
        return self._width

    @property
    def width_hint(self):
        if self._width:
            return None
        return self._width_hint or 1

    @property
    def markup(self):
        return self._markup

    @property
    def i18n(self):
        return self._i18n


class BasicTableSpec(object):
    def __init__(self, parent, column_specs, name='UNK', show_column_header=True,
                 dedup_keys=False, row_height=90, row_spacing=10,
                 font_size='42sp', font_bold=False, font_name=None, font_context=None):
        self._name = name
        self._parent = parent
        self._column_specs = column_specs
        for colspec in self._column_specs:
            colspec.parent = self
        self._show_column_header = show_column_header
        self._row_height = row_height
        self._row_spacing = row_spacing
        self._font_size = font_size
        self._font_bold = font_bold
        self._dedup_keys = dedup_keys
        self._font_name = font_name
        self._font_context = font_context

    def install(self):
        pass

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @property
    def font_name(self):
        return self._font_name

    @property
    def font_context(self):
        return self._font_context

    @property
    def font_size(self):
        return self._font_size

    @property
    def font_bold(self):
        return self._font_bold

    @property
    def font_params(self):
        params = {
            'font_context': self.font_context,
            'font_name': self.font_name,
            'font_size': self.font_size,
            'bold': self.font_bold,
        }
        return {k: v for k, v in params.items() if v is not None}

    @property
    def show_column_header(self) -> bool:
        return self._show_column_header

    @property
    def column_specs(self) -> list:
        return self._column_specs

    @property
    def row_height(self):
        return self._row_height

    @property
    def row_spacing(self):
        return self._row_spacing

    @property
    def dedup_keys(self):
        return self._dedup_keys





from kivy.uix.boxlayout import BoxLayout

from kivy.graphics import Color
from kivy.graphics import Rectangle
from kivy.properties import ListProperty

from itertools import chain
from kivy.graphics.texture import Texture


def color_set_alpha(color, alpha):
    cl = list(color[:3])
    cl.append(alpha)
    return cl


class GuiPalette(object):
    def __init__(self, background, foreground, color_1, color_2):
        self.background = background
        self.foreground = foreground
        self.color_1 = color_1
        self.color_2 = color_2

    def __repr__(self):
        rv = "<{0}>".format(self.__class__.__name__)
        rv += "\n - background : {}".format(self.background)
        rv += "\n - foreground : {}".format(self.foreground)
        rv += "\n - color_1 : {}".format(self.color_1)
        rv += "\n - color_2 : {}".format(self.color_2)
        return rv


class Gradient(object):

    @staticmethod
    def horizontal(*args):
        texture = Texture.create(size=(len(args), 1), colorfmt='rgba')
        buf = bytes([int(v * 255) for v in chain(*args)])  # flattens
        texture.blit_buffer(buf, colorfmt='rgba', bufferfmt='ubyte')
        return texture

    @staticmethod
    def vertical(*args):
        texture = Texture.create(size=(1, len(args)), colorfmt='rgba')
        buf = bytes([int(v * 255) for v in chain(*args)])  # flattens
        texture.blit_buffer(buf, colorfmt='rgba', bufferfmt='ubyte')
        return texture


class BackgroundColorMixin(object):
    bgcolor = ListProperty([1, 1, 1, 1])

    def __init__(self, bgcolor=None):
        self.bgcolor = bgcolor or [1, 1, 1, 1]
        self._render_bg()
        self.bind(size=self._render_bg, pos=self._render_bg)
        self.bind(bgcolor=self._render_bg)

    def _render_bg(self, *args):
        self.canvas.before.clear()
        with self.canvas.before:
            Color(*self.bgcolor)
            Rectangle(pos=self.pos, size=self.size)

    def make_transparent(self):
        self.bgcolor = color_set_alpha(self.bgcolor, 0)

    def make_opaque(self):
        self.bgcolor = color_set_alpha(self.bgcolor, 1)


class ColorBoxLayout(BackgroundColorMixin, BoxLayout):
    def __init__(self, **kwargs):
        bgcolor = kwargs.pop('bgcolor')
        BoxLayout.__init__(self, **kwargs)
        BackgroundColorMixin.__init__(self, bgcolor=bgcolor)





import os
import atexit
import shutil
import tempfile
from appdirs import user_cache_dir
from twisted.internet import reactor
from .structure import BaseGuiStructureMixin
from .widgets.colors import GuiPalette


class BaseMixin(object):
    _appname = 'iotnode'
    _app_root = os.path.abspath(os.path.dirname(__file__))
    _app_resources = os.path.join(_app_root, 'resources')

    def __init__(self, *args, **kwargs):
        self._reactor = kwargs.pop('reactor', reactor)
        self._cache_dir = None
        self._db_dir = None
        self._temp_dir = None
        super(BaseMixin, self).__init__(*args, **kwargs)

    @property
    def app_resources(self):
        return self._app_resources

    def install(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def _deferred_error_passthrough(self, failure):
        return failure

    def _deferred_error_swallow(self, failure):
        return

    @property
    def reactor(self):
        return self._reactor

    @property
    def appname(self):
        return self._appname

    @property
    def cache_dir(self):
        if not self._cache_dir:
            self._cache_dir = user_cache_dir(self.appname)
            os.makedirs(self._cache_dir, exist_ok=True)
        return self._cache_dir

    @property
    def db_dir(self):
        if not self._db_dir:
            self._db_dir = os.path.join(self.cache_dir, 'db')
            os.makedirs(self._db_dir, exist_ok=True)
        return self._db_dir

    @property
    def temp_dir(self):
        if not self._temp_dir:
            self._temp_dir = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, self._temp_dir)
        return self._temp_dir


class BaseGuiMixin(BaseGuiStructureMixin):
    _palette = GuiPalette(
        background=(0x00 / 255, 0x00 / 255, 0x00 / 255),
        foreground=(0xff / 255, 0xff / 255, 0xff / 255),
        color_1=(0xff / 255, 0xff / 255, 0xff / 255),
        color_2=(0xff / 255, 0xff / 255, 0xff / 255)
    )

    def gui_setup(self):
        pass

    @property
    def gui_color_1(self):
        return self._palette.color_1

    @property
    def gui_color_2(self):
        return self._palette.color_2

    @property
    def gui_color_foreground(self):
        return self._palette.foreground

    @property
    def gui_color_background(self):
        return self._palette.background





from . import BaseShellMixin


class WifiNetworkInfoMixin(BaseShellMixin):

    @property
    def wifi_ssid(self):
        def _handle_result(result):
            return result.strip()
        d = self._shell_execute(['iwgetid', '-s'], _handle_result)
        return d


class NetworkInfoMixin(WifiNetworkInfoMixin):

    @property
    def network_info(self):
        return self.wifi_ssid





import gc
from collections import OrderedDict

from kivy.animation import Animation
from kivy.uix.relativelayout import RelativeLayout

from ebs.iot.linuxnode.animations.composite import CompositeAnimationManager

from .translatable import TranslatableRenderableTable


class AnimatedTable(TranslatableRenderableTable):
    def __init__(self, node, spec):
        self._current_entries = OrderedDict()
        self._animations = CompositeAnimationManager()
        self._animation_layer = None
        self._animation_lock = False
        super(AnimatedTable, self).__init__(node, spec)

    @property
    def active_entries(self):
        raise NotImplementedError

    def _build_title(self):
        raise NotImplementedError

    @property
    def animation_layer(self):
        if not self._animation_layer:
            self._animation_layer = RelativeLayout()
            self._gui_table_entries_container.add_widget(self._animation_layer)
        return self._animation_layer

    def _calculate_animation_distance(self, idx_source=None, idx_destination=None):
        if idx_source is None or idx_destination is None:
            ad_x = self._gui_table_entries.width
            ad_y = 0
        else:
            ad_x = 0
            ad_y = (idx_source - idx_destination) * self.row_pitch
        return ad_x, ad_y

    def _entry_pos(self, idx):
        return 0, self._gui_table_entries.height - (idx + 1) * self.row_pitch + 10

    def _entry_size(self):
        return self._gui_table_entries.width, self.row_pitch - 10

    def _delay_exit(self, idx):
        return Animation(d=0.1 * idx)

    def _delay_entry(self, idx):
        return Animation(d=0.1 * idx + 0.8)

    def _detach_entry(self, entry_bin):
        if entry_bin.parent == self._gui_table_entries:
            pos = entry_bin.pos
            self._gui_table_entries.remove_widget(entry_bin)
            entry_bin.pos = pos
            self.animation_layer.add_widget(entry_bin)

    def _detach_entries(self, entries):
        for tags, entry_bin in entries.items():
            self._detach_entry(entry_bin)
        self.log.debug("Detached {0} Entries, Remaining T={1},A={2} Entries"
                       "".format(len(entries),
                                 len(self._gui_table_entries.children),
                                 len(self.animation_layer.children)))

    def _exit_animation(self, idx):
        x, y = self._entry_pos(idx)
        dx, dy = self._calculate_animation_distance(idx, None)
        animation = Animation(x=x + dx, y=y + dy,
                              t='in_cubic', duration=1.)
        return animation

    def _destroy_dangling_entities(self):
        for instance in self.animation_layer.children:
            if hasattr(instance, 'stop'):
                instance.stop()
        self.animation_layer.clear_widgets()

    def _preposition_entry(self, idx, entry_bin):
        x, y = self._entry_pos(idx)
        dx, dy = self._calculate_animation_distance(None, idx)
        entry_bin.pos = x - dx, y - dy
        self.animation_layer.add_widget(entry_bin)

    def _position_entry(self, idx, entry_bin):
        entry_bin.pos = self._entry_pos(idx)
        self.animation_layer.add_widget(entry_bin)

    def _entry_animation(self, idx):
        px, py = self._entry_pos(idx)
        animation = Animation(x=px, y=py, t='out_cubic', duration=1.)
        return animation

    def _transfer_animation(self, idx):
        px, py = self._entry_pos(idx)
        animation = Animation(x=px, y=py, t='in_out_cubic', duration=1.)
        return animation

    def _still_animation(self):
        animation = Animation(d=1.)
        return animation

    def _attach_entry(self, entry_bin):
        self.animation_layer.remove_widget(entry_bin)
        self._gui_table_entries.add_widget(entry_bin)

    def _attach_entries(self, entries):
        for tags, entry_bin in entries.items():
            self._attach_entry(entry_bin)
        self.log.debug("Attached {0} Entries, Remaining T={1},A={2} Entries"
                       "".format(len(entries),
                                 len(self._gui_table_entries.children),
                                 len(self.animation_layer.children)))

    def _entries_change_handler(self, _, value):
        pass

    def _alternate_fallback_handler(self, _, entries):
        if self._gui_fallback:
            if not len(entries):
                if self._gui_table in self._gui_table_container.children:
                    self._gui_table_container.remove_widget(self._gui_table)
                if self._gui_fallback not in self._gui_table_container.children:
                    self._gui_table_container.add_widget(self._gui_fallback)
            else:
                if self._gui_fallback in self.gui_table_container.children:
                    self._gui_table_container.remove_widget(self._gui_fallback)
                if self._gui_table not in self.gui_table_container.children:
                    self._gui_table_container.add_widget(self._gui_table)

    def _build_entries(self, entries):
        rv = OrderedDict()
        for entry in entries:
            tags = [getattr(entry, x) for x in self.spec.dedup_keys]
            tags.append(self.language)
            tags = tuple(tags)
            entry_bin = entry.build(self.palette)
            entry_bin.size_hint = (None, None)
            entry_bin.size = self._entry_size()
            rv[tags] = entry_bin
        return rv

    def redraw_entries(self, entries):
        if self._animation_lock:
            self.log.warn("Animation lock is active! Force breaking.")

        self.log.debug("Redrawing Table, Got {0} Entries".format(len(entries)))
        self._alternate_fallback_handler(None, entries)
        self._detach_entries(self._current_entries)
        new_entries = self._build_entries(entries)
        deferred_tags = OrderedDict()

        self._animations.clear()

        for idx, (tags, entry_bin) in enumerate(self._current_entries.items()):
            if tags in new_entries.keys():
                deferred_tags[tags] = (idx, entry_bin)
                continue
            animation = self._delay_exit(idx) + self._exit_animation(idx)
            self._animations.add(animation, entry_bin)

        for idx, (tags, entry_bin) in enumerate(new_entries.items()):
            if tags in deferred_tags.keys():
                oidx, oentry_bin = deferred_tags[tags]
                self.animation_layer.remove_widget(oentry_bin)
                self._position_entry(oidx, entry_bin)
                if oidx < idx:
                    animation = self._delay_entry(idx) + self._transfer_animation(idx)
                else:
                    animation = self._transfer_animation(idx)
                self._animations.add(animation, entry_bin)
                continue
            self._preposition_entry(idx, entry_bin)
            animation = self._delay_entry(idx) + self._entry_animation(idx)
            self._animations.add(animation, entry_bin)

        def _finish():
            self._animation_lock = False
            self._attach_entries(self._current_entries)
            self._destroy_dangling_entities()
            gc.collect()

        self._current_entries = new_entries

        if len(self._animations):
            self._animations.when_done(_finish)
            self._animation_lock = True
            self._animations.start()
        else:
            _finish()





import re
from kivy.uix.label import Label
from kivy.uix.widget import Widget
from kivy.uix.scrollview import ScrollView

from kivy.core.window import Window
from kivy.animation import Animation
from kivy.properties import NumericProperty

from .colors import ColorBoxLayout


class MarqueeLabel(ScrollView):
    spacer_width = NumericProperty(None)

    def __init__(self, **kwargs):
        bgcolor = kwargs.pop('bgcolor')
        ScrollView.__init__(self, bar_width=0)

        self._layout = ColorBoxLayout(size_hint_x=None,
                                      bgcolor=bgcolor,
                                      orientation='horizontal')
        self._mainlabels = []
        self._lspacer = Widget(size_hint_x=None, width=Window.width)
        self._rspacer = Widget(size_hint_x=None, width=Window.width)
        self.add_widget(self._layout)

        text = kwargs.pop('text')
        self._label_params = kwargs
        self.text = text

    def update_widths(self):
        width = self._lspacer.width + \
                self._rspacer.width + \
                sum([x.width for x in self._mainlabels])
        self._layout.width = width
        self.width = width

    def _set_spacer_width(self, _, size):
        self._lspacer.width = size[0]
        self._rspacer.width = size[0]
        self.update_widths()

    def on_parent(self, _, parent):
        if parent:
            parent.bind(size=self._set_spacer_width)

    @property
    def text(self):
        return ' '.join([x.text for x in self._mainlabels])

    def _set_mainlabel_width(self, l, size):
        l.width = size[0]
        self.update_widths()

    @text.setter
    def text(self, value):
        self.remove_widget(self._layout)

        texts = split_string(value, 64)
        self._layout.clear_widgets()
        self._layout.add_widget(self._lspacer)

        self._mainlabels = []
        for t in texts:
            l = Label(text=t, size_hint_x=None,
                      text_size=(None, None), **self._label_params)
            self._layout.add_widget(l)
            l.bind(texture_size=self._set_mainlabel_width)
            l.texture_update()
            self._mainlabels.append(l)

        self._layout.add_widget(self._rspacer)
        self.add_widget(self._layout)
        self.update_widths()

    def start(self, loop=True):
        speed = 75
        scroll_distance = self._layout.width - self._lspacer.width
        duration = scroll_distance / speed
        # print("Scroll was", self.scroll_x)
        self.scroll_x = 0
        # print("Using duration", duration)
        self._animation = Animation(scroll_x=1, duration=duration)
        if loop:
            self._animation.bind(on_complete=self._check_complete)
            pass
        self._animation.start(self)

    def _check_complete(self, animation, instance):
        # print(instance.scroll_x)
        # print(self._mainlabels[-1].x + self._mainlabels[-1].width)
        if instance.scroll_x > 0.95:
            # print("Repeating")
            self._animation.unbind(on_scroll=self._check_complete)
            animation.stop(self)
            self.start()

    def stop(self):
        self._animation.unbind(on_scroll=self._check_complete)
        self._animation.stop(self)
        self.clear_widgets()


def split_string(text, limit):
    words = re.split('(\W)', text)
    if max(map(len, words)) > limit:
        raise ValueError("limit is too small")
    result = []
    cpart = ''
    for word in words:
        if len(word) > limit - len(cpart):
            result.append(cpart)
            cpart = word
        else:
            cpart += word
    if cpart:
        result.append(cpart)
    return result





import os

from ..log import NodeLoggingMixin
from ..http import HttpClientMixin
from ..basemixin import BaseGuiMixin

from ..widgets.image import BleedImage
from ..widgets.labels import ColorLabel

from ..widgets.image import StandardImage
from ..widgets.labels import SelfScalingLabel
from ..widgets.colors import ColorBoxLayout


class ModularApiEngineManagerMixin(HttpClientMixin, NodeLoggingMixin):
    def __init__(self, *args, **kwargs):
        super(ModularApiEngineManagerMixin, self).__init__(*args, **kwargs)
        self._api_engines = []
        self._api_primary = None

    def modapi_install(self, engine, primary=False):
        self.log.info("Installing Modular API Engine {0}".format(engine))
        self._api_engines.append(engine)
        if primary:
            self._api_primary = engine

    def modapi_activate(self):
        for engine in self._api_engines:
            self.log.info("Starting Modular API Engine {0}".format(engine))
            engine.start()

    def modapi_engine(self, name):
        for engine in self._api_engines:
            if engine.name == name:
                return engine

    def modapi_stop(self):
        for engine in self._api_engines:
            self.log.info("Stopping Modular API Engine {0}".format(engine))
            engine.stop()

    def start(self):
        super(ModularApiEngineManagerMixin, self).start()
        self.modapi_activate()

    def stop(self):
        self.modapi_stop()
        super(ModularApiEngineManagerMixin, self).stop()


class ModularApiEngineManagerGuiMixin(ModularApiEngineManagerMixin, BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        super(ModularApiEngineManagerGuiMixin, self).__init__(*args, **kwargs)

        self._api_internet_link = None
        self._api_internet_link_indicator = None

        self._api_internet_connected = False
        self._api_internet_indicator = None

        self._api_connection_status = {}
        self._api_connection_indicators = {}

    @property
    def modapi_internet_link_indicator(self):
        if not self._api_internet_link_indicator:
            params = {'bgcolor': (0xff / 255., 0x00 / 255., 0x00 / 255., 0.3),
                      'color': [1, 1, 1, 1]}
            self._api_internet_link_indicator = ColorLabel(
                text=self._api_internet_link, size_hint=(None, None),
                height=50, font_size='14sp',
                valign='middle', halign='center', **params
            )

            def _set_label_width(_, texture_size):
                self._api_internet_link_indicator.width = texture_size[0] + 20

            self._api_internet_link_indicator.bind(texture_size=_set_label_width)
        return self._api_internet_link_indicator

    def _modapi_internet_link_indicator_show(self, duration=5):
        _ = self.modapi_internet_link_indicator
        if not self._api_internet_link_indicator.parent:
            self.gui_notification_stack.add_widget(self._api_internet_link_indicator)
            self.gui_notification_update()
        if duration:
            self.reactor.callLater(duration, self._modapi_internet_link_indicator_clear)

    def _modapi_internet_link_indicator_clear(self):
        if self._api_internet_link_indicator and self._api_internet_link_indicator.parent:
            self.gui_notification_stack.remove_widget(self._api_internet_link_indicator)
            self.gui_notification_update()
        self._api_internet_link_indicator = None

    @property
    def modapi_internet_indicator(self):
        if not self._api_internet_indicator:
            _root = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
            _source = os.path.join(_root, 'images', 'no-internet.png')
            self._api_internet_indicator = BleedImage(
                source=_source, pos_hint={'left': 1},
                size_hint=(None, None), height=50, width=50,
                bgcolor=(0xff / 255., 0x00 / 255., 0x00 / 255., 0.3),
            )
        return self._api_internet_indicator

    def _modapi_internet_indicator_show(self):
        if not self.modapi_internet_indicator.parent:
            self.gui_notification_row.add_widget(self.modapi_internet_indicator)
            self.gui_notification_update()

    def _modapi_internet_indicator_clear(self):
        if self.modapi_internet_indicator.parent:
            self.modapi_internet_indicator.parent.remove_widget(self.modapi_internet_indicator)
            self.gui_notification_update()

    def modapi_connection_indicator(self, prefix):
        if prefix not in self._api_connection_indicators.keys():
            _root = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
            source = os.path.join(_root, 'images', 'no-server.png')
            indicator = ColorBoxLayout(
                pos_hint={'left': 1}, orientation='vertical', padding=(0, 0, 0, 5),
                size_hint=(None, None), height=70, width=50, spacing=0,
                bgcolor=(0xff / 255., 0x00 / 255., 0x00 / 255., 0.3),
            )
            indicator.add_widget(
                StandardImage(source=source, size_hint=(1, None), height=50)
            )
            indicator.add_widget(
                SelfScalingLabel(text=prefix,
                                 size_hint=(1, None), height=15)
            )

            self._api_connection_indicators[prefix] = indicator
        return self._api_connection_indicators[prefix]

    def _modapi_connection_indicator_show(self, prefix):
        if not self.modapi_connection_indicator(prefix).parent:
            self.gui_notification_row.add_widget(self.modapi_connection_indicator(prefix))
            self.gui_notification_update()

    def _modapi_connection_indicator_clear(self, prefix):
        if self.modapi_connection_indicator(prefix).parent:
            self.modapi_connection_indicator(prefix).parent.remove_widget(self.modapi_connection_indicator(prefix))
            self.gui_notification_update()

    def modapi_signal_internet_link(self, value, prefix):
        self._api_internet_link = value
        self._modapi_internet_link_indicator_show()

    def modapi_signal_internet_connected(self, value, prefix):
        if not value:
            self._modapi_internet_indicator_show()
        else:
            self._modapi_internet_indicator_clear()
        self._api_internet_connected = value

    def modapi_signal_api_connected(self, value, prefix):
        if not value:
            self._modapi_connection_indicator_show(prefix)
        else:
            self._modapi_connection_indicator_clear(prefix)
        self._api_connection_status[prefix] = value





import os
import glob
import atexit
import shutil
import tempfile

from threading import Thread
from pdf2image import convert_from_path
from kivy.properties import StringProperty
from kivy.properties import BooleanProperty
from kivy.properties import NumericProperty
from kivy.uix.floatlayout import FloatLayout
from kivy.clock import Clock

from .gallery import ImageGallery


def generate_pdf_images(source, target, callback):

    if not os.path.exists(source):
        raise FileNotFoundError(source)

    if os.path.exists(target):
        if callback:
            callback()
        return

    os.makedirs(target, exist_ok=True)

    def _namegen():
        while True:
            yield 'page'

    convert_from_path(source, fmt='png', dpi=100,
                      output_folder=target,
                      output_file=_namegen())

    if callback:
        callback()


class PDFPlayer(FloatLayout):
    source = StringProperty()
    loop = BooleanProperty(True)
    interval = NumericProperty(10)

    def __init__(self, source, loop=True, temp_dir=None,
                 exit_retrace=False, **kwargs):
        super(PDFPlayer, self).__init__(**kwargs)

        self._gallery = ImageGallery(parent_layout=self,
                                     animation_vector=(-1, 0),
                                     exit_retrace=exit_retrace)
        self._gallery.transition = 'in_out_expo'

        self._current_page = -1
        self._task = None
        self._cancelled = False
        self._pages = []

        if not temp_dir:
            temp_dir = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, temp_dir)
        self._temp_dir = temp_dir

        self.bind(source=self._load_source)
        self.bind(interval=self.start)

        self.loop = loop
        self.source = source

    @property
    def num_pages(self):
        return len(self._pages)

    @property
    def pages_dir(self):
        name = os.path.splitext(os.path.basename(self.source))[0]
        return os.path.join(self._temp_dir, name)

    def _generate_images(self, callback):
        generate_pdf_images(self.source, self.pages_dir, callback)

    def _load_source(self, *_):
        if not self.source:
            return

        if not os.path.exists(self.source):
            raise FileNotFoundError(self.source)

        if not os.path.exists(self.pages_dir):
            Thread(target=self._generate_images,
                   args=[self._start_display]).start()
        else:
            self._start_display()

    def _start_display(self):
        if self._cancelled:
            return

        def _sort_key(filepath):
            fname = os.path.splitext(os.path.basename(filepath))[0]
            return int(fname.split('-')[1])

        self._pages = sorted(glob.glob(os.path.join(self.pages_dir, '*.png')),
                             key=_sort_key)
        self.start()

    def stop(self):
        self._cancelled = True
        if self._task:
            self._task.cancel()

    def _next_page(self):
        if self._current_page < len(self._pages) - 1:
            return self._current_page + 1
        else:
            return 0

    def step(self, *_):
        last_page = self._current_page
        self._current_page = self._next_page()
        if last_page == self._current_page:
            return
        self._gallery.current = self._pages[self._current_page]

    def start(self):
        self.stop()
        if not self._pages:
            return
        self._task = Clock.schedule_interval(self.step, self.interval)
        self.step()



import setuptools

_requires = [
    'six',
    'setuptools-scm',
    'appdirs',
    'Cython',
    'twisted[tls]',
    'SQLAlchemy',
    'kivy>=1.11.1',
    'kivy-garden',
    'arrow',
    'raspi_system',

    # Node Id
    'netifaces',

    # BleedImage
    'colorthief',
    
    # HTTP Client
    'treq',

    # Event Manager
    'cached_property',
    'pqueue',

    # Browser
    'selenium',

    # RPi Mediaplayer
    'omxplayer-wrapper',

    # PdfPlayer
    'pdf2image',

    # Text and Localization
    'hoshi',
]

setuptools.setup(
    name='ebs-iot-linuxnode',
    url='',

    author='Chintalagiri Shashank',
    author_email='shashank.chintalagiri@gmail.com',

    description='',
    long_description='',

    packages=setuptools.find_packages(),
    package_dir={'ebs.iot.linuxnode': 'ebs/iot/linuxnode'},
    package_data={'ebs.iot.linuxnode': ['images/no-internet.png',
                                        'images/no-server.png']},

    install_requires=_requires,

    setup_requires=['setuptools_scm'],
    use_scm_version=True,

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
    ],
)




from twisted.internet.defer import Deferred

from .widgets.colors import color_set_alpha
from .widgets.marquee import MarqueeLabel

from .basemixin import BaseGuiMixin
from .config import ConfigMixin


class MarqueeInterrupted(Exception):
    pass


class MarqueeBusy(Exception):
    def __init__(self, now_playing, collision_count):
        self.now_playing = now_playing
        self.collision_count = collision_count

    def __repr__(self):
        return "<MarqueeBusy Now Playing {0}" \
               "".format(self.now_playing)


class MarqueeGuiMixin(ConfigMixin, BaseGuiMixin):
    _gui_marquee_bgcolor = None
    _gui_marquee_color = None

    def __init__(self, *args, **kwargs):
        super(MarqueeGuiMixin, self).__init__(*args, **kwargs)
        self._gui_marquee = None
        self._marquee_text = None
        self._marquee_deferred = None
        self._marquee_end_call = None
        self._marquee_collision_count = 0

    def marquee_show(self):
        self.gui_footer_show()

    def marquee_hide(self):
        self.gui_footer_hide()

    def marquee_play(self, text, duration=None, loop=True):
        if self._marquee_deferred:
            self._marquee_collision_count += 1
            if self._marquee_collision_count > 30:
                self.marquee_stop(forced=True)
            raise MarqueeBusy(self._marquee_text,
                              self._marquee_collision_count)
        self._marquee_collision_count = 0
        self.gui_marquee.text = text
        self.marquee_show()

        if duration:
            self._gui_marquee.start(loop=loop)
            self._marquee_end_call = self.reactor.callLater(duration, self.marquee_stop)
        else:
            self._gui_marquee.start(callback=self.marquee_stop)
        self._marquee_deferred = Deferred()
        return self._marquee_deferred

    def marquee_stop(self, forced=False):
        self._gui_marquee.stop()
        self.log.info("End Offset by {0} collisions."
                      "".format(self._marquee_collision_count))
        self._marquee_collision_count = 0
        self.marquee_hide()

        if self._marquee_end_call and self._marquee_end_call.active():
            self._marquee_end_call.cancel()

        if self._marquee_deferred:
            self._marquee_deferred.callback(forced)
            self._marquee_deferred = None

    @property
    def gui_marquee(self):
        if not self._gui_marquee:
            params = {'bgcolor': (self._gui_marquee_bgcolor or
                                  color_set_alpha(self.gui_color_2, 0.6)),
                      'color': [1, 1, 1, 1],
                      'font_size': '32sp'}
            if self.config.text_font_name:
                params['font_name'] = self.config.text_font_name
            self._gui_marquee = MarqueeLabel(text='Marquee Text', **params)

            self.gui_footer.add_widget(self._gui_marquee)
            self.marquee_hide()
        return self._gui_marquee

    def gui_setup(self):
        super(MarqueeGuiMixin, self).gui_setup()
        _ = self.gui_marquee





from twisted.internet import threads
from twisted.internet.defer import Deferred
from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options

from .basemixin import BaseMixin
from .config import ConfigMixin
from .background import OverlayWindowGuiMixin


class BrowserManager(object):
    def __init__(self, node, bmid):
        self._bmid = bmid
        self._node = node
        self._target = None
        self._browser = None

    @property
    def options(self):
        start_page = self._node.config.browser_default_url
        self._node.log.debug("Starting browser with {url}", url=start_page)
        chrome_options = Options()
        chrome_options.add_argument("--disable-extensions")
        chrome_options.add_argument("--app={0}".format(start_page))
        # if self._node.config.http_proxy_enabled:
        #     chrome_options.add_argument(
        #         "--proxy={0}".format(self._node.config.http_proxy_url)
        #     )
        return chrome_options

    @property
    def browser(self):
        if not self._browser:
            d = threads.deferToThread(Chrome, options=self.options)

            def _attach_browser(browser):
                self._browser = browser
            d.addCallback(_attach_browser)
            self._browser = d
        return self._browser

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        if isinstance(self._browser, Deferred):
            def _postponed(*args):
                self.target = value
            self._browser.addCallback(_postponed)
        else:
            self._node.reactor.callInThread(self.browser.get, value)
            self._target = value

    def clear(self):
        self.target = 'about:blank'

    def set_geometry(self, x, y, width, height):
        self._node.log.debug("Setting browser geometry : "
                             "({x}, {y}), ({width}, {height})",
                             x=x, y=y, width=width, height=height)
        self.browser.set_window_rect(x, y, width, height)

    def start(self):
        return self.browser

    def terminate(self):
        if not self._browser or isinstance(self._browser, Deferred):
            return
        self._browser.close()
        self._browser.quit()


class BrowserMixin(ConfigMixin, BaseMixin):
    def __init__(self, *args, **kwargs):
        self._browser_managers = {}
        super(BrowserMixin, self).__init__(*args, **kwargs)

    def browser_manager(self, bmid):
        if bmid not in self._browser_managers.keys():
            self.log.info("Initializing browser manager {bmid}", bmid=bmid)
            self._browser_managers[bmid] = BrowserManager(self, bmid)
        return self._browser_managers[bmid]

    @property
    def browser(self):
        return self.browser_manager(0)

    def browser_start(self):
        maybe_deferred = self.browser_manager(0).start()
        if isinstance(maybe_deferred, Deferred):
            maybe_deferred.addCallback(lambda _: self.gui_browser_show())
        else:
            self.gui_browser_show()

    def browser_stop(self):
        if self.config.browser_show_default:
            self.browser.target = self.config.browser_default_url
        else:
            self.gui_browser_hide()
            self.browser_manager(0).clear()

    def gui_browser_show(self):
        pass

    def gui_browser_hide(self):
        pass

    def start(self):
        super(BrowserMixin, self).start()
        if self.config.browser_show_default:
            self._reactor.callLater(1, self.browser_start)

    def stop(self):
        for bmid in self._browser_managers.keys():
            self._browser_managers[bmid].terminate()
        super(BrowserMixin, self).stop()


class BrowserGuiMixin(OverlayWindowGuiMixin):
    def __init__(self, *args, **kwargs):
        self._browser_visible = False
        super(BrowserGuiMixin, self).__init__(*args, **kwargs)

    @property
    def gui_browser_container(self):
        return self.gui_sidebar_right

    def gui_browser_show(self):
        if not self.overlay_mode:
            self.overlay_mode = True
        self._browser_visible = True
        self.gui_sidebar_right_show('browser')
        self.browser.set_geometry(
            self.gui_browser_container.x, self.gui_browser_container.y,
            self.gui_browser_container.width, self.gui_browser_container.height
        )

    def gui_browser_hide(self):
        self._browser_visible = False
        self.gui_sidebar_right_hide('browser')

    def gui_setup(self):
        super(BrowserGuiMixin, self).gui_setup()




import os
from twisted import logger
from twisted.internet.defer import succeed
from twisted.internet.task import LoopingCall

from ..common import HTTPError
from .primitives import ApiPersistentActionQueue


class ModularApiEngineBase(object):
    _prefix = ""
    _api_probe = None
    _api_tasks = []
    _api_reconnect_frequency = 30

    def __init__(self, actual):
        self._log = None
        self._actual = actual
        self._api_reconnect_task = None
        self._api_engine_active = False
        self._api_endpoint_connected = None
        self._api_queue = ApiPersistentActionQueue(self)

    @property
    def name(self):
        return self._prefix

    """ Proxy to Core Engine """

    @property
    def cache_dir(self):
        return os.path.join(self._actual.cache_dir, self._prefix)

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="modapi.{0}".format(self._prefix), source=self)
        return self._log

    """ API Connection Status Primitives """
    @property
    def api_endpoint_connected(self):
        return self._api_endpoint_connected

    @api_endpoint_connected.setter
    def api_endpoint_connected(self, value):
        self._actual.modapi_signal_api_connected(value, self._prefix)
        self._api_endpoint_connected = value

    """ API Task Management """
    @property
    def api_tasks(self):
        return self._api_tasks

    def install_task(self, task, period):
        self._api_tasks.append((task, period))

    def _api_start_all_tasks(self, _):
        for task, period in self.api_tasks:
            t = getattr(self, task)
            if not t.running:
                self.log.info("Starting {task} with period {period}",
                              task=task, period=period)
                t.start(period)
        return succeed(True)

    def _api_stop_all_tasks(self, _):
        for task, _ in self._api_tasks:
            t = getattr(self, task)
            if t.running:
                self.log.info("Stopping {task}", task=task)
                t.stop()
        return succeed(True)

    """ API Connection Management """
    @property
    def api_reconnect_task(self):
        if self._api_reconnect_task is None:
            self._api_reconnect_task = LoopingCall(self.api_engine_activate)
        return self._api_reconnect_task

    def api_engine_activate(self):
        self.log.debug("Attempting to activate {0} API engine.".format(self._prefix))

        d = getattr(self, self._api_probe)()

        def _made_connection(_):
            self.log.debug("Made connection")
            self.api_endpoint_connected = True
            self._api_engine_active = True
            if self.api_reconnect_task.running:
                self.api_reconnect_task.stop()
            self.log.info("Triggering process of {0} API persistent queue".format(self._prefix))
            self._api_queue.process()
            return

        def _enter_reconnection_cycle(failure):
            self.log.error("Can't connect to {0} API endpoint".format(self._prefix))
            self.log.failure("Connection Failure : ", failure=failure)
            self.api_endpoint_connected = False
            if not self.api_reconnect_task.running:
                self.api_engine_reconnect()
            return failure

        d.addCallbacks(
            _made_connection,
            _enter_reconnection_cycle
        )

        def _error_handler(failure):
            if self.api_reconnect_task.running:
                return
            else:
                print("Returning failure ", failure)
                return failure

        d.addCallbacks(
            self._api_start_all_tasks,
            _error_handler
        )
        return d

    def api_engine_reconnect(self):
        if self._api_engine_active:
            self.api_endpoint_connected = False
            self.log.info("Lost connection to {0} API server. Attempting to reconnect."
                          "".format(self._prefix))
        self._api_engine_active = False
        if not self.api_reconnect_task.running:
            self._api_stop_all_tasks(True)
            self.api_reconnect_task.start(self._api_reconnect_frequency)

    def api_engine_stop(self):
        self._api_engine_active = False
        for task, _ in self._api_tasks:
            if getattr(self, task).running:
                getattr(self, task).stop()
        if self.api_reconnect_task.running:
            self.api_reconnect_task.stop()

    @property
    def api_engine_active(self):
        return self._api_engine_active

    def start(self):
        self.api_engine_activate()

    def stop(self):
        self.api_engine_stop()


class ModularHttpApiEngine(ModularApiEngineBase):
    _api_baseurl = ''
    _api_headers = {}

    def __init__(self, actual, config=None):
        super(ModularHttpApiEngine, self).__init__(actual)
        self._config = config
        self._api_token = None
        self._internet_connected = False
        self._internet_link = None

    """ Proxy to Core Engine """
    @property
    def http_get(self):
        return self._actual.http_get

    @property
    def http_post(self):
        return self._actual.http_post

    @property
    def network_info(self):
        return self._actual.network_info

    @property
    def config(self):
        if self._config:
            return self._config
        else:
            return self._actual.config

    """ Network Status Primitives """

    # TODO Consider moving the core network status primitives
    #  entirely into the manager instead

    @property
    def internet_connected(self):
        return self._internet_connected

    @internet_connected.setter
    def internet_connected(self, value):
        self._actual.modapi_signal_internet_connected(value, self._prefix)
        self._internet_connected = value

    @property
    def internet_link(self):
        return self._internet_link

    @internet_link.setter
    def internet_link(self, value):
        self._actual.modapi_signal_internet_link(value, self._prefix)
        self._internet_link = value

    """ API Engine Management """
    def api_engine_activate(self):
        # Probe for internet
        d = self.http_get('https://www.google.com')

        def _get_internet_info(maybe_failure):
            ld = self.network_info

            def _set_internet_link(l):
                if l:
                    self.internet_link = l.decode('utf-8')
            ld.addCallback(_set_internet_link)
            return maybe_failure
        d.addBoth(_get_internet_info)

        def _made_connection(_):
            self.log.debug("Have Internet Connection")
            self.internet_connected = True

        def _enter_reconnection_cycle(failure):
            self.log.error("No Internet!")
            self.internet_connected = False
            if not self.api_reconnect_task.running:
                self.api_engine_reconnect()
            return failure

        d.addCallbacks(
            _made_connection,
            _enter_reconnection_cycle
        )

        def _error_handler(failure):
            if self.api_reconnect_task.running:
                return
            else:
                return failure

        d.addCallbacks(
            lambda _: ModularApiEngineBase.api_engine_activate(self),
            _error_handler
        )
        return d

    @property
    def api_token(self):
        raise NotImplementedError

    def api_token_reset(self):
        raise NotImplementedError

    """ Core HTTP API Executor """
    def _api_execute(self, ep, request_builder, response_handler):
        url = "{0}/{1}".format(self.api_url, ep)
        d = self.api_token
        d.addCallback(request_builder)

        def _get_response(req: dict):
            self.log.debug("Executing API Request to {url} \n"
                           "   with content '{content}'\n"
                           "   and headers '{headers}'", 
                           url=url, content=req, headers=self._api_headers)
            params = req.pop('_query', [])
            request_structure = {
                'json': req,
                'params': params,
            }
            request_structure = {k: v for k, v in request_structure.items() if v}
            r = self.http_post(url, timeout=120,
                               headers=self._api_headers,
                               **request_structure)
            return r
        d.addCallback(_get_response)

        def _error_handler(failure):
            self.log.failure("Attempting to handle API Error", failure=failure)
            if isinstance(failure.value, HTTPError) and \
                    failure.value.response.code == 403:
                self.log.debug("Encountered 403 Error. Attempting API Token Reset.")
                self.api_token_reset()
            if not self.api_reconnect_task.running:
                self.log.debug("Starting API Reconnect Task")
                self.api_engine_reconnect()
            return failure
        d.addCallbacks(response_handler, _error_handler)
        return d

    @property
    def api_url(self):
        if self._api_baseurl.startswith('config'):
            cft = self._api_baseurl.split(':')[1]
            return getattr(self.config, cft)
        return self._api_baseurl

    def start(self):
        super(ModularHttpApiEngine, self).start()

    def stop(self):
        super(ModularHttpApiEngine, self).stop()






from math import floor, ceil

from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.relativelayout import RelativeLayout

from ebs.iot.linuxnode.tables.basic import BasicTableEntry, BasicTable
from ebs.iot.linuxnode.widgets.colors import ColorBoxLayout
from ebs.iot.linuxnode.widgets.image import BleedImage
from ebs.iot.linuxnode.widgets.labels import SelfScalingColorLabel


class BasicRenderableTableEntry(BasicTableEntry):
    def __init__(self, data):
        super(BasicRenderableTableEntry, self).__init__(data)

    def build(self, palette=None):
        if not palette:
            palette = self.parent.palette

        _gui_entry = ColorBoxLayout(orientation='horizontal', spacing=10,
                                    bgcolor=palette.grid_background,
                                    size_hint=(1, None),
                                    height=self.parent.spec.row_height)

        _gui_entry.add_widget(BoxLayout(size_hint=(None, None), width=20,
                                        height=self.parent.spec.row_height))

        for colspec in self.parent.spec.column_specs:
            l_font_params = self.parent.spec.font_params
            l_font_params.update({
                'bold': colspec.font_bold
            })
            kwargs = dict(
                text=self.parent.preprocess(getattr(self, colspec.accessor), colspec),
                bgcolor=palette.cell_background,
                color=palette.cell_foreground,
                size_hint=(colspec.width_hint, None),
                height=self.parent.spec.row_height,
                valign='middle', halign=colspec.halign,
                padding_x=15, width=colspec.width,
                markup=colspec.markup,
                **l_font_params
            )
            label = SelfScalingColorLabel(
                **{k: v for k, v in kwargs.items() if v is not None}
            )
            label.bind(size=label.setter('text_size'))
            _gui_entry.add_widget(label)

        _gui_entry.add_widget(BoxLayout(size_hint=(None, None), width=20,
                                        height=self.parent.spec.row_height))
        return _gui_entry


class BasicRenderableTable(BasicTable):
    def __init__(self, node, spec):
        self._gui_table_container = None
        self._gui_table = None
        self._gui_table_title = None
        self._gui_table_header = None
        self._gui_table_entries_container = None
        self._gui_table_entries = None
        self._fallback_image = None
        self._gui_fallback = None
        self._palette = None
        super(BasicRenderableTable, self).__init__(node, spec)

    @property
    def palette(self):
        return self._palette

    @palette.setter
    def palette(self, value):
        self._palette = value

    def preprocess(self, value, spec=None):
        return str(value)

    @property
    def gui_table_container(self):
        if not self._gui_table_container:
            self._gui_table_container = BoxLayout()
        return self._gui_table_container

    def _build_header(self, palette):
        _gui_table_header = ColorBoxLayout(orientation='horizontal',
                                           spacing=self.spec.row_spacing,
                                           bgcolor=palette.grid_background,
                                           size_hint=(1, None),
                                           height=self.spec.row_height)

        _gui_table_header.add_widget(BoxLayout(size_hint=(None, None), width=20,
                                               height=self.spec.row_height))

        l_font_params = self.spec.font_params
        l_font_params.update({
            'bold': True
        })

        for colspec in self.spec.column_specs:
            kwargs = dict(
                text=self.preprocess(colspec.title),
                bgcolor=palette.header_cell_background,
                color=palette.header_cell_foreground,
                size_hint=(colspec.width_hint, None),
                height=self.spec.row_height,
                valign='middle', halign='center',
                padding_x=15, width=colspec.width,
                **l_font_params
            )
            label = SelfScalingColorLabel(
                **{k: v for k, v in kwargs.items() if v is not None}
            )
            label.bind(size=label.setter('text_size'))
            _gui_table_header.add_widget(label)

        _gui_table_header.add_widget(BoxLayout(size_hint=(None, None), width=20,
                                               height=self.spec.row_height))
        return _gui_table_header

    @property
    def row_pitch(self):
        return self.spec.row_height + self.spec.row_spacing

    @property
    def entries_per_page(self):
        if self._gui_table in self._gui_table_container.children:
            return floor(self._gui_table_entries.height / self.row_pitch) - 1
        else:
            # TODO This is a race condition during GUI construction
            return 6

    @property
    def active_entries(self):
        raise NotImplementedError

    @property
    def total_pages(self):
        return ceil(len(self.active_entries) / self.entries_per_page)

    def page_entities(self, page):
        e_tot = len(self.active_entries)
        p_tot = ceil(e_tot / self.entries_per_page)
        if page == p_tot - 1:
            s = max(0, e_tot - self.entries_per_page)
            e = e_tot
        else:
            s = page * self.entries_per_page
            e = s + self.entries_per_page
        return self.active_entries[s:e]

    def _build_title(self):
        raise NotImplementedError

    @property
    def fallback_image(self):
        return self._fallback_image

    @fallback_image.setter
    def fallback_image(self, value):
        self._fallback_image = value

    def _entries_change_handler(self, _, value):
        if not self._gui_fallback:
            return
        if not len(value):
            if self._gui_table in self._gui_table_container.children:
                self._gui_table_container.remove_widget(self._gui_table)
            if self._gui_fallback not in self._gui_table_container.children:
                self._gui_table_container.add_widget(self._gui_fallback)
        else:
            if self._gui_fallback in self.gui_table_container.children:
                self._gui_table_container.remove_widget(self._gui_fallback)
            if self._gui_table not in self.gui_table_container.children:
                self._gui_table_container.add_widget(self._gui_table)

    def build(self, entries):
        if self.fallback_image:
            self._gui_fallback = BleedImage(
                source=self.fallback_image,
                allow_stretch=True,
                keep_ratio=True,
                bgcolor='auto'
            )
        self._gui_table = ColorBoxLayout(orientation='vertical',
                                         bgcolor=self.palette.grid_background)

        self._gui_table_entries = GridLayout(cols=1, size_hint=(1, None),
                                             spacing=self.spec.row_spacing)
        self._gui_table_entries.bind(children=self._entries_change_handler)

        # # This is required to overlay an animation layer.
        self._gui_table_entries_container = RelativeLayout()
        # self._gui_table_entries_container.bind(size=self._gui_table_entries.setter("size"))
        # self._gui_table_entries_container.bind(pos=self._gui_table_entries.setter("pos"))
        self._gui_table_entries_container.add_widget(self._gui_table_entries)

        _reserved_height = 0

        try:
            self._gui_table_title = self._build_title()
            self._gui_table.add_widget(self._gui_table_title)
            _reserved_height += self._gui_table_title.height
        except NotImplementedError:
            pass

        if self.spec.show_column_header:
            self._gui_table_header = self._build_header(self.palette)
            self._gui_table.add_widget(self._gui_table_header)
            _reserved_height += self._gui_table_header.height

        self._gui_table.add_widget(BoxLayout(size_hint=(1, None),
                                             height=self.spec.row_spacing))
        _reserved_height += self.spec.row_spacing

        def _table_entries_height(_, value):
            self._gui_table_entries.height = value - _reserved_height
        self._gui_table.bind(height=_table_entries_height)

        self.gui_table_container.clear_widgets()
        self.gui_table_container.add_widget(self._gui_table)
        self._gui_table.add_widget(self._gui_table_entries_container)

        self.redraw_entries(entries)
        if hasattr(self, '_alternate_fallback_handler'):
            handler = getattr(self, '_alternate_fallback_handler')
        else:
            handler = self._entries_change_handler
        handler(self._gui_table_entries, self._gui_table_entries.children)
        return self.gui_table_container

    def redraw_entries(self, entries):
        self.log.debug("Redrawing Table, Got {0} Entries".format(len(entries)))
        if hasattr(self, '_alternate_fallback_handler'):
            self._alternate_fallback_handler(None, entries)
        self._gui_table_entries.clear_widgets()
        for entry in entries:
            self._gui_table_entries.add_widget(entry.build(self.palette))
